/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef MESH_H
#define MESH_H

#include "Brush.h"
#include "Math.h"
#include "Optimized.h"
#include "Polygon3D.h"
#include "Texture.h"
#include <map>
#include <set>
#include <sigc++/sigc++.h>
#include <string>
#include <vector>

namespace SharpConstruct
{
	namespace Interface
	{
		class ViewGL;
	}
	class ExternalRender;
}

namespace SharpConstruct
{
/** Stores and manipulates vertex and polygon data.
 *
 * Provides storage for vertex and polygon data, as well as data used to
 * display the mesh, and data relating the vertex and polygon data. Active data
 * is kept involving the area of the mesh currently being edited, and the
 * edit configuration. Functions are provided for file input/output, interactive
 * editing, deformations, and retrieval of mesh information.
 *
 * Note that in the context of this class, "selection" refers to the sphere of
 * points centered at _center and within the diameter of _diameter. "Area"
 * refers to whatever mirror of the selection is being worked on. The area may
 * be the selection when not mirrored over any axis.
**/
class Mesh
{
	friend class SharpConstruct::Interface::ViewGL;
	friend class SharpConstruct::ExternalRender;
	
public:
	Mesh();
	~Mesh()
	{}

	// Basic file input/output control
	void Load( std::string filename );
	void Save( std::string filename );

	// Create model data
	void BecomePlane( int, int );
	void BecomeSphere( unsigned );
	void BecomeCube( unsigned );

	// Brushing Operations
	void DisplaceAreas();
	void StartGrab();
	void ContinueGrab( float, float );
	void EndGrab();
	void SmoothAreas();
	void PinchAreas();
	void MaskAreas();

	// Deformations
	void Move( float, float, float );
	void Resize( float, float, float );
	void Deflate( float v );
	void Smooth( float );
	void Flood( float );
	void Subdivide( bool smooth );
	void Optimize( float );
	int RemoveDuplicates();
	void Mirror( bool, bool, bool );
	void FlipNormals();
	void Unify();
	void ToTriangles();

	//void SetTexture( Magick::Image * );

	// Mesh information access
	sigc::signal< void >& Changed();
	std::string FileName() const;
	bool Saved() const;
	std::string GetModelStatisticsString();
	float GetDiameter();
	void CopyEssentials( Mesh & );
	void ClearExtraData();

	void RecalculateAllNormals();
	
	static SharpConstruct::Brush *MainBrush;

	void PerformanceTest( std::string test );
private:
	// File input/output
	void _load_obj( std::string );
	void _save_obj( std::string );
	void _save_wrl( std::string );

	Optimized::Point3D _vertex_neighbor_loc_average( unsigned v );
	Color _vertex_neighbor_clr_average( unsigned v );

	void _displace_area( EditData e );
	void _smooth_area( EditData e );
	void _continue_grab( std::vector< unsigned > &, std::vector< float > &,
	                     EditData e, SharpConstruct::Optimized::Normal3D );
	void _pinch_area( EditData e );
	Math::PolarPoint _relative_2d_location( EditData& e, Optimized::Point3D x, float len );
	EditData _get_mirror( bool x, bool y, bool z );

	void _calculate_bounding_sphere();
	void _find_active_vertices( std::vector< unsigned > &,
	                            std::vector< float > &,
				    EditData e );
	void _find_active_grab_vertices( std::vector< unsigned > &,
	                            std::vector< float > &,
				    EditData e );

	class ActiveVertexData
	{
	public:
		ActiveVertexData( unsigned c, std::set< unsigned >& chk, std::vector< unsigned >& indices,
				  std::vector< float >& dists, Optimized::Point3D& a )
			: Current( c ), Checked( chk ), VertexIndices( indices ),
			  VertexDistances( dists ), Area( a )
		{}

		unsigned Current;
		std::set< unsigned >& Checked;
		std::vector< unsigned >& VertexIndices;
		std::vector< float >& VertexDistances;
		Optimized::Point3D& Area;
	};
	void _is_vertex_active( ActiveVertexData& d );
	Optimized::Point3D _get_mirror_point( bool, bool, bool );

	// Calculate normals
	Optimized::Point3D _calculate_area_normal();
	void _update_polygon_normal( int );
	void _recalculate_modified_normals();
	
	// Object data
	void _change_operation();
	sigc::signal< void > _changed;
	bool _saved;
	std::string _filename;

	Texture _txtr;
	
	unsigned _poly_count() const;
	void _clear_polys();
	Optimized::Point3DVector _vertex_locations;
	Optimized::Point3DVector _vertex_normals;
	std::vector< SharpConstruct::Color > _vertex_colors;
	std::vector< SharpConstruct::UV > _uvs;
	std::vector< Triangle > _triangles;
	std::vector< Quad > _quads;
	// Stores the normals of each polygon
	Optimized::Point3DVector _triangle_normals;
	Optimized::Point3DVector _quad_normals;
	// Stores the polygons that use each vertex
	std::vector< PolygonUsers > _vertex_users;
	
	void _recalculate_damaged();
	std::list< unsigned > _damaged_triangles;
	std::list< unsigned > _damaged_quads;
	
	void _recalculate_vertex_users();
	std::map< unsigned, unsigned > _triangle_users( unsigned v1, unsigned v2 );
	std::map< unsigned, unsigned > _quad_users( unsigned v1, unsigned v2 );
	std::set< unsigned > _connected_vertices( unsigned v );
	Optimized::Point3D _polygon_center( Triangle& t );
	Optimized::Point3D _polygon_center( Quad& q );
	
	// Active data
	float _selection_radius;
	static Optimized::Point3D _current_selection;
	unsigned _selected_vertex[ 8 ];
	std::vector< unsigned > _active_vertex_indices;
	std::vector< float > _active_vertex_distances;
	static Optimized::Normal3D _view_up_normal;
	static Optimized::Normal3D _view_right_normal;
	unsigned _lowest_index, _highest_index;
	std::vector< unsigned > _modified_vertex_indices;
	std::vector< unsigned > _grab_vertices[ 8 ];
	std::vector< float > _grab_vertex_distances[ 8 ];
	//static Magick::Image *_texture;
	
	// Bounding Sphere
	float _diameter;
	Optimized::Point3D _center;
};
}
#endif // MESH_H
