/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef OPTIMIZED_H
#define OPTIMIZED_H

#include <xmmintrin.h>
#include <cmath>
#include <cstdlib>
#include <new>

namespace SharpConstruct
{
	namespace Optimized
	{
		const int OPT_SIZE = 16;
		class Point3D
		{
		public:
			inline Point3D() : _proxy( false )
			{
				alloc();
			}
			inline Point3D( const Point3D& in ) : _proxy( false )
			{
				alloc();
				*_data = *in._data;
			}
			inline Point3D( const __m128& in ) : _proxy( false )
			{
				alloc();
				*_data = in;
			}
			inline Point3D( const float x, const float y, const float z ) : _proxy( false )
			{
				alloc();
				*_data = _mm_set_ps( 0, z, y, x );
			}
			inline ~Point3D()
			{
				if( !_proxy )
				{
					free( _data );
					_data = 0;
				}
			}
			void CalculatePlaneNormal( __m128* p1,  __m128* p2,
			                           __m128* p3 );
			void CalculatePlaneUnormal( __m128* p1,  __m128* p2,
			                            __m128* p3 );
			inline void Normalize()
			{
				register __m128 tmp = *_data;
				register __m128 length;
				register __m128 shuff;
				// Square all the components
				tmp = _mm_mul_ps( tmp, tmp );
				// Next six lines are to put X + Y + Z into length
				shuff = _mm_shuffle_ps( tmp, tmp, _MM_SHUFFLE( 1, 2, 3, 0 ) );
				length = shuff;
				shuff = _mm_shuffle_ps( tmp, tmp, _MM_SHUFFLE( 2, 3, 0, 1 ) );
				length = _mm_add_ss( length, shuff );
				shuff = _mm_shuffle_ps( tmp, tmp, _MM_SHUFFLE( 3, 0, 1, 2 ) );
				length = _mm_add_ss( length, shuff );
				// Take the reciprocal square root
				length = _mm_rsqrt_ss( length );
				// Move the rsqrt into all four positions
				length = _mm_shuffle_ps( length, length, _MM_SHUFFLE( 0, 0, 0, 0 ) );
				// Multiply it by data
				*_data = _mm_mul_ps( *_data, length );
			}
			inline float Distance( const Point3D& in )
			{
				register __m128 p1 = *_data, shuff, dis;
				p1 = _mm_sub_ps( p1, *in._data );
				p1 = _mm_mul_ps( p1, p1 );
				shuff = _mm_shuffle_ps( p1, p1, _MM_SHUFFLE( 1, 2, 3, 0 ) );
				dis = shuff;
				shuff = _mm_shuffle_ps( p1, p1, _MM_SHUFFLE( 2, 3, 0, 1 ) );
				dis = _mm_add_ss( dis, shuff );
				shuff = _mm_shuffle_ps( p1, p1, _MM_SHUFFLE( 3, 0, 1, 2 ) );
				dis = _mm_add_ss( dis, shuff );
				dis = _mm_rsqrt_ss( dis );
				dis = _mm_rcp_ss( dis );
				// Is there a faster way to do this?
				//return ((float*)&dis)[0];
				// Yes: _mm_store_ss
				float d;
				_mm_store_ss( &d, dis );
				return d;
			}
			// Not a true distance!
			void FastDistance( const Point3D& in, float& d );

			void Midpoint( const Point3D& a, const Point3D& b );
			inline void CopyTo( float* loc )
			{
				_mm_storeu_ps( loc, *_data );
			}
			inline void CopyToAligned( float* loc )
			{
				_mm_store_ps( loc, *_data );
			}
			inline float HorizAdd() const
			{
				float a = 0;
				register __m128 t = *_data;
				t = _mm_add_ps( t, _mm_shuffle_ps( t, t, _MM_SHUFFLE( 2, 1, 0, 3 ) ) );
				t = _mm_add_ps( t, _mm_shuffle_ps( t, t, _MM_SHUFFLE( 2, 2, 2, 2 ) ) );
				_mm_store_ss( &a, t );
				return a;
			}
			inline void Zero()
			{
				*_data = _mm_setzero_ps();
			}
			inline Point3D& Abs()
			{
				register __m128 v( _mm_set_ps1( -0.0 ) );
				*_data = _mm_andnot_ps( v, *_data );
				return *this;
			}
			inline float& X()
			{
				return ( ( float* )_data )[ 0 ];
			}
			inline float& Y()
			{
				return ( ( float* )_data )[ 1 ];
			}
			inline float& Z()
			{
				return ( ( float* )_data )[ 2 ];
			}
			inline float& W()
			{
				return ( ( float* )_data )[ 3 ];
			}
			inline const float& X() const
			{
				return ( ( float* )_data )[ 0 ];
			}
			inline const float& Y() const
			{
				return ( ( float* )_data )[ 1 ];
			}
			inline const float& Z() const
			{
				return ( ( float* )_data )[ 2 ];
			}
			inline Point3D& operator=( const Point3D& in )
			{
				*_data = *in._data;
				return *this;
			}			
			inline operator __m128() const
			{
				return *_data;
			}
			inline bool operator==( const Point3D& p ) const
			{
				return X() == p.X() && Y() == p.Y() && Z() == p.Z();
			}
			inline Point3D operator+( const Point3D& in ) const
			{
				return Point3D( _mm_add_ps( *_data, *in._data ) );
			}
			inline Point3D operator-( const Point3D& in )
			{
				return Point3D( _mm_sub_ps( *_data, *in._data ) );
			}
			inline Point3D operator*( const Point3D& in ) const
			{
				return Point3D( _mm_mul_ps( *_data, *in._data ) );
			}
			inline Point3D operator*( const float in )
			{
				return Point3D( X() * in, Y() * in, Z() * in );
			}
			inline Point3D operator/( const float in )
			{
				return Point3D( X() / in, Y() / in, Z() / in );
			}
			inline void operator+=( const Point3D& in )
			{
				*_data = _mm_add_ps( *_data, *in._data );
			}
			inline void operator-=( const Point3D& in )
			{
				*_data = _mm_sub_ps( *_data, *in._data );
			}
			inline void operator/=( const float in )
			{
				register __m128 d = *_data, div = _mm_set1_ps( in );
				d = _mm_div_ps( d, div );
				*_data = d;
			}
			inline __m128* RawData() const
			{
				return _data;
			}
		protected:
			__m128* _data;
			bool _proxy;
		
			// Kinda kludgy, I guess. Anyway, this ctor is only for
			// Point3DProxy. Has a dummy parameter to distinguish it.
			inline Point3D( int ) : _proxy( true )
			{}
		private:
			void alloc()
			{
				void* mem;
				posix_memalign( &mem, __alignof( __m128 ), sizeof( __m128 ) );
				_data = new( mem ) __m128;
			}
		};
		
		typedef Point3D Normal3D;
		
		class Point3DProxy : public Point3D
		{
		public:
			inline Point3DProxy( __m128* in ) : Point3D( 1 )
			{
				Set( in );
			}
			inline void Set( __m128* in )
			{
				_data = in;
			}
			inline Point3DProxy& operator=( const Point3D& in )
			{
				*_data = *in.RawData();
				return *this;
			}
		};
		
		class Point3DVector
		{
		public:		
			inline Point3DVector() : _data( 0 ), _size( 0 ), _real_size( 0 )
			{}
			inline Point3DVector( const Point3DVector& in ) : _data( 0 ), _size( 0 ), _real_size( 0 )
			{
				resize( in._size );
				for( int i = 0; i < _size; ++i )
					_data[ i ] = in._data[ i ];
			}
			inline ~Point3DVector()
			{
				clear();
			}

			void NormalizeAll();

			inline unsigned size() const
			{
				return _size;
			}
			inline void clear()
			{
				free( _data );
				_data = 0;
				_size = 0;
				_real_size = 0;
			}
			inline void resize( int size )
			{
				if( size > _real_size )
				{
					Point3DVector tmp;
					if( size > 0 )
						tmp = *this;
					
					clear();

					_real_size = static_cast< int >(
						pow( 2, round( log( size ) / log( 2 ) ) ) );
					if( _real_size < size )
						_real_size *= 2;
					
					void* mem;
					posix_memalign( &mem, __alignof( __m128 ),
									_real_size * sizeof( __m128 ) );
					_data = ( __m128* )mem;
					for( int i = 0; i < _real_size; ++i )
						new( _data + sizeof( __m128 ) * i ) __m128;
					
					if( size > 0 )
					{
						for( int i = 0; i < tmp._size; ++i )
							_data[ i ] = tmp._data[ i ];
					}
				}

				_size = size;
			}
			inline void push_back( const Point3D& in )
			{
				resize( _size + 1 );
				_data[ _size - 1 ] = in;
			}
			inline Point3DProxy operator[]( int i )
			{				
				//_proxy.Set( &_data[ i ] );
				//return _proxy;
				return Point3DProxy( &_data[ i ] );
			}
			inline Point3DVector& operator=( const Point3DVector& in )
			{
				resize( in._size );
				for( int i = 0; i < _size; ++i )
					_data[ i ] = in._data[ i ];
				
				return ( *this );
			}
			inline __m128* RawData() const
			{
				return _data;
			}
		private:		
			__m128* _data;
			int _size;
			int _real_size;
		};
	}
}

#endif // OPTIMIZED_H
