/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "BrushEditor.h"
#include "FileDialog.h"
#include "MessageDialog.h"
#include "TopWindow.h"
#include "config.h"
#include "prefix.h"
#include <GL/gl.h>
#include <dirent.h>
#include <sys/stat.h>

using namespace SharpConstruct::Interface;

TextureSwatch::TextureSwatch( std::string path ) :
	Toggle( 32, 32, "" )
{
	_txtr.Load( path );
	Icon().SetTexture( _txtr );

	MouseUpSignal.connect( sigc::mem_fun( *this,
					      &TextureSwatch::_click ) );

	GetTooltip().SetMessage( FileDialog::FilePart( path ) );
}

SharpConstruct::Texture& TextureSwatch::Txtr()
{
	return _txtr;
}

sigc::signal< void, TextureSwatch* >& TextureSwatch::ClickSignal()
{
	return _click_signal;
}

void TextureSwatch::_click()
{
	ClickSignal()( this );
	SetValue( true );
}

BrushEditor::BrushEditor() :
	_GRD_SZ( 3 ),
	_add( "Add" ),
	_remove( "Remove" ),
	_scroll( 0, 0, 20, THMB_SZ * _GRD_SZ ),
	_cur( 0 )
{
	_add.GetTooltip().SetMessage( "Load an image to be used as the brush alpha.",
				      "The image must be square and power of two (usually "
				      "128 or 256 pixels.) If it is not, it can be "
				      "converted automatically." );
	_remove.GetTooltip().SetMessage( "Remove the currently selected alpha.",
					 "The image on the disk is not modified." );

	_scroll.SetMax( 1 );

	_cols.resize( _GRD_SZ );
	for( unsigned i = 0; i < _cols.size(); i++ )
	{
		_cols[ i ].SetW( THMB_SZ );
		_all.PushBack( _cols[ i ] );
	}

	_controls.PushBack( _add );
	_controls.PushBack( _remove );
	PushBack( _controls );
	PushBack( _all );
	PushBack( _scroll );

	_add.MouseUpSignal.connect( ShowOpenDialogSignal().make_slot() );
	_remove.MouseUpSignal.connect( sigc::mem_fun( *this,
						      &BrushEditor::_remove_handler ) );
	_scroll.ChangedSignal.connect( sigc::mem_fun( *this,
						      &BrushEditor::_scroll_handler ) );

	SetVisible( false );

	_load_default_brushset();
}

BrushEditor::~BrushEditor()
{
	for( unsigned i = 0; i < _cols.size(); i++ )
		for( Iterator j = _cols[ i ].Begin();
		     j != _cols[ i ].End(); j++ )
			delete *j;
}

bool BrushEditor::AddAlpha( std::string path, bool show_msgs )
{
	Texture test;
	if( !test.Load( path ) )
	{
		if( show_msgs )
			MessageDialog::ModalMessage( TopWindow::Display(), "Image error",
						     "File " + path + "\n cannot be loaded; this format "
						     "is not supported.", "Cancel" );
		return false;
	}
	if( test.W() != test.H() )
	{
		std::string answr = MessageDialog::ModalMessage( TopWindow::Display(), "Alpha requirements",
								 "File " + path + "\ncannot be loaded because it is rectangular, "
								 "which is not currently supported.\n\n"
								 "Automatically correct texture? (This change will not be saved.)",
								 "Apply Correction", "Cancel" );
		if( answr == "Cancel" )
			return false;
	}

	_alphas.push_back( new TextureSwatch( path ) );
	_alphas.back()->Txtr().AutoCorrect();
	_cur = _alphas.back();
	_cur->ClickSignal().connect( sigc::mem_fun( *this,
						    &BrushEditor::_swatch_click_handler )  );

	int m = 0;
	for( unsigned i = 1; i < _cols.size(); i++ )
		if( _cols[ i ].List< Widget* >::Size() <
		    _cols[ m ].List< Widget* >::Size() )
			m = i;

	_cols[ m ].PushBack( _cur );
	_cur->MouseUpSignal();
	_all.Reorganize();

	_scroll.SetMax( _cols[ 0 ].List< Widget* >::Size() - 2 );

	return true;
}

SharpConstruct::Texture& BrushEditor::CurrentAlpha()
{
	return _cur->Txtr();
}

sigc::signal< void >& BrushEditor::ShowOpenDialogSignal()
{
	return _show_open_dialog_signal;
}

sigc::signal< void, SharpConstruct::Texture& >& BrushEditor::AlphaChangedSignal()
{
	return _alpha_changed_signal;
}

void BrushEditor::_draw()
{
	glEnable( GL_SCISSOR_TEST );
	glMatrixMode( GL_PROJECTION_MATRIX );
	glScissor( X(), TopWindow::Display().H() - Y() - H(), W(), H() );
	glMatrixMode( GL_MODELVIEW_MATRIX );
	HorizontalGroup::_draw();
	glDisable( GL_SCISSOR_TEST );
}

void BrushEditor::_load_default_brushset()
{
	std::string def_path( BR_DATADIR( "/sharpconstruct/brushes/" ) );
	AddAlpha( def_path + "Default.png", false );
	
	TextureSwatch* frst = _cur;

	int count;
	struct dirent** files;
	struct stat statinfo;
	std::string tmp;
	
	count = scandir( def_path.c_str(), &files, 0, alphasort );
	for( int i = 2; i < count; ++i ) // Start at two so we skip . and ..
	{
		tmp = def_path + files[ i ]->d_name;
		stat( tmp.c_str(), &statinfo );
		if( !S_ISDIR( statinfo.st_mode ) && strcmp( files[ i ]->d_name, "Default.png" ) ) 			AddAlpha( tmp, false );
	}
	free( files );

	_cur = frst;
	frst->MouseUpSignal();
}

void BrushEditor::_swatch_click_handler( TextureSwatch* t )
{
	_cur = t;

	for( unsigned i = 0; i < _cols.size(); i++ )
		for( Iterator j = _cols[ i ].Begin();
		     j != _cols[ i ].End(); j++ )
			static_cast< TextureSwatch* >( *j )->SetValue( false );

	AlphaChangedSignal()( t->Txtr() );
}

void BrushEditor::_remove_handler()
{
	if( _cols[ 0 ].List< Widget* >::Size() <= 1 &&
	    _cols[ 1 ].List< Widget* >::Size() == 0 )
		return;
	std::vector< VerticalGroup > cpy( _cols.size() );
	unsigned cc = 0;
	Iterator p;
	TextureSwatch* fsave = 0;
	unsigned i = 0;
	int found = 0; // 0=store cur swatch, 1=Set when del swatch found, store next one, 2=don't store anymore
	while( 1 )
	{
		if( _cols[ i ].List< Widget* >::Size() )
		{
			p = _cols[ i ].Begin();

			if( ( *p ) != _cur )
			{
				cpy[ cc++ ].PushBack( *p );
				if( cc >= _cols.size() )
					cc = 0;
				if( found < 2 )
					fsave = static_cast< TextureSwatch* >( *p );
				if( found == 1 )
					found = 2;
			}
			else
				found = 1;
			_cols[ i ].List< Widget* >::Erase( p );
		}
		else
			break;
		i++;
		if( i >= _cols.size() )
			i = 0;
	}

	if( fsave )
	{
		fsave->SetValue( true );
		_cur = fsave;
	}

	for( unsigned i = 0; i < _cols.size(); i++ )
	{
		_cols[ i ].Assign( cpy[ i ] );
		_cols[ i ].Reorganize();
	}
	_all.Reorganize();

	_cur->MouseUpSignal();
}

void BrushEditor::_scroll_handler()
{
	_all.SetY( Y() - _scroll.Value() * THMB_SZ );
}
