/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "DeformDialog.h"
#include "MainWindow.h"
#include "Utilities.h"
#include <sys/time.h>

using namespace SharpConstruct::Interface;

/**
 * This file has become more complex than I would like; it seems that, with all
 * most of the actual deforming being done in the Mesh class, this one should be
 * pretty simple. The reason it isn't is the previewing stuff. And once that was
 * in, even more complicated code was added to toggle previews. A cleaner design
 * would be nice though.
**/

DeformDialog::DeformDialog( MainWindow& mw )
: Dialog( "Deform" ),
  _parent( mw ),
  _axes( LARGEST_CHILD, 0, 4 ), _xaxis( "X" ), _yaxis( "Y" ), _zaxis( "Z" ),
  _slide_defs( LARGEST_CHILD, 4, 4 ),
  _preview( "Preview" ),
  _move( "Move" ),
  _scale( "Scale" ),
  _deflate( "Deflate" ),
  _smooth( "Smooth" ),
  _flood( "Flood" ),
  _subdivide( "Subdivide" ),
  _sub_smooth( "Smooth" ),
  _optimize( "Optimize" ),
  _mirror( "Mirror" ),
  _flip_normals( "Flip Normals" ),
  _unify( "Unify" ),
  _to_triangles( "To Triangles" ),
  _pt( "PerfTest" )
{
	// Yes, I know...
	_xaxis.SetW( 67 );
	_xaxis.SetValue( true );
	_yaxis.SetValue( true );
	_zaxis.SetValue( true );
	
	_move.SetProperties( -100, 100, 1, 0, false );
	_scale.SetProperties( -100, 100, 1, 0, false );
	_deflate.SetProperties( -100, 100, 1, 0, false );
	_smooth.SetProperties( -100, 100, 1, 0, false );
	_flood.SetProperties( 0, 100, 1, 0, false );
	
	_axes.PushBack( _xaxis );
	_axes.PushBack( _yaxis );
	_axes.PushBack( _zaxis );
	_slide_defs.PushBack( _preview );
	_slide_defs.PushBack( _move );
	_slide_defs.PushBack( _scale );
	_slide_defs.PushBack( _deflate );
	_slide_defs.PushBack( _smooth );
	_slide_defs.PushBack( _flood );
	_subdivide.SetW( 153 ); // Grr.
	_sub_group.PushBack( _subdivide );
	_sub_group.PushBack( _sub_smooth );
	PushBack( _axes );
	PushBack( _slide_defs );
	PushBack( _sub_group );
	//PushBack( _optimize );
	PushBack( _mirror );
	PushBack( _flip_normals );
	PushBack( _unify );
	PushBack( _to_triangles );
	//PushBack( _pt );

	_preview.GetTooltip().SetMessage( "Keep updating the model when the sliders below are changed.",
					  "This option is usually best left on, but it may be useful to turn it off "
					  "if the model is very high-poly, when updating may be too slow." );
	_move.GetTooltip().SetMessage( "Move the model along the selected axes." );
	_scale.GetTooltip().SetMessage( "Scale the model along the selected axes." );
	_deflate.GetTooltip().SetMessage( "Pull each vertex inward along its normal. Negative values push the vertex outwards." );
	_smooth.GetTooltip().SetMessage( "Average each vertex with its neighbors; smooths sharp edges in the model. "
					 "Negative values sharpen the model." );
	_flood.GetTooltip().SetMessage( "Flood the entire model with the currently selected color." );
	_subdivide.GetTooltip().SetMessage( "Replace each polygon with several smaller polygons.",
					    "Each quad will become four new quads, while each "
					    "triangle will become three new quads. This deformation"
					    "is affected by the Smooth toggle at the left." );
	_sub_smooth.GetTooltip().SetMessage( "Smooth vertices during a Subdivide deformation." );
	_mirror.GetTooltip().SetMessage( "Flip the model across the selected axes." );
	_flip_normals.GetTooltip().SetMessage( "Flip all polygon normals." );
	_unify.GetTooltip().SetMessage( "Move the model so that its center is at the origin." );
	_to_triangles.GetTooltip().SetMessage( "Convert all quads into triangles." );

	_preview.SetValue( true );
	_preview.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_preview_handler ) );
	_preview_handler();
	
	_subdivide.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_subdivide ) );
	_optimize.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_optimize ) );
	_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_mirror ) );
	_flip_normals.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_flip_normals ) );
	_unify.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_unify ) );
	_to_triangles.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_to_triangles ) );

	_pt.MouseUpSignal.connect( sigc::mem_fun( *this, &DeformDialog::_cb_pt ) );
}

void DeformDialog::_make_copy()
{
	_mesh_copy.CopyEssentials( _parent.GetMeshHistory().GetCurrentMesh() );
}
void DeformDialog::_revert_to_copy( float )
{
	_parent.GetMeshHistory().GetCurrentMesh().CopyEssentials( _mesh_copy );
}
void DeformDialog::_transfer( float )
{
	Mesh tmp;
	tmp.CopyEssentials( _parent.GetMeshHistory().GetCurrentMesh() );
	_parent.GetMeshHistory().GetCurrentMesh().CopyEssentials( _mesh_copy );
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().CopyEssentials( tmp );	
	_parent.View().Refresh();
}

void DeformDialog::_prepare_change( float )
{
	_parent.GetMeshHistory().AddMesh();
}

void DeformDialog::_preview_handler()
{
	std::vector< Slider* > s;
	s.push_back( &_move );
	s.push_back( &_scale );
	s.push_back( &_deflate );
	s.push_back( &_smooth );
	s.push_back( &_flood );
	
	for( unsigned i = 0; i < s.size(); i++ )
	{
		// Clear all the signals
		s[ i ]->BeginSignal().slots().erase(
			s[ i ]->BeginSignal().slots().begin(),
			s[ i ]->BeginSignal().slots().end() );
		s[ i ]->ChangedSignal().slots().erase(
			s[ i ]->ChangedSignal().slots().begin(),
			s[ i ]->ChangedSignal().slots().end() );
		s[ i ]->SetSignal().slots().erase(
			s[ i ]->SetSignal().slots().begin(),
			s[ i ]->SetSignal().slots().end() );
		if( _preview.Value() )
		{
			s[ i ]->BeginSignal().connect( sigc::mem_fun( *this, &DeformDialog::_make_copy ) );
			s[ i ]->ChangedSignal().connect( sigc::mem_fun( *this, &DeformDialog::_revert_to_copy ) );
			s[ i ]->SetSignal().connect( sigc::mem_fun( *this, &DeformDialog::_transfer ) );
		}
		else
			s[ i ]->SetSignal().connect( sigc::mem_fun( *this, &DeformDialog::_prepare_change ) );
	}
	
	if( _preview.Value() )
	{
		_move.ChangedSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_move ) );
		_scale.ChangedSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_scale ) );
		_deflate.ChangedSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_deflate ) );
		_smooth.ChangedSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_smooth ) );
		_flood.ChangedSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_flood ) );
	}
	else
	{
		_move.SetSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_move ) );
		_scale.SetSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_scale ) );
		_deflate.SetSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_deflate ) );
		_smooth.SetSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_smooth ) );
		_flood.SetSignal().connect( sigc::mem_fun( *this, &DeformDialog::_cb_flood ) );
	}
}

void DeformDialog::_cb_move( float v )
{
	float scaled_value = v / 100;
	scaled_value *= _parent.GetMeshHistory().GetCurrentMesh().GetDiameter();
	_parent.GetMeshHistory().GetCurrentMesh().Move(
		scaled_value * _xaxis.Value(), scaled_value * _yaxis.Value(),
		scaled_value * _zaxis.Value() );
}
void DeformDialog::_cb_scale( float v )
{
	float clipped_value = v / 50;
	if( clipped_value > 0 )
		clipped_value++;
	else
		clipped_value--;
	_parent.GetMeshHistory().GetCurrentMesh().Resize( clipped_value * _xaxis.Value(),
	                                  clipped_value * _yaxis.Value(),
	                                  clipped_value * _zaxis.Value() );
}
void DeformDialog::_cb_deflate( float v )
{
	_parent.GetMeshHistory().GetCurrentMesh().Deflate( v / 200 );
}
void DeformDialog::_cb_smooth( float v )
{
	_parent.GetMeshHistory().GetCurrentMesh().Smooth( v / 100 );
}
void DeformDialog::_cb_flood( float v )
{
	_parent.GetMeshHistory().GetCurrentMesh().Flood( v / 100 );
}
void DeformDialog::_cb_subdivide()
{
	_parent.SetStatusText( "Subdividing..." );
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Subdivide( _sub_smooth.Value() );
	_parent.GetMeshHistory().GetCurrentMesh().RecalculateAllNormals();
	_parent.SetStatusText( "Subdivide complete; " + _parent.GetMeshHistory().
		GetCurrentMesh().GetModelStatisticsString() );
}
void DeformDialog::_cb_optimize()
{
	/*_parent.SetStatusText( "Optimizing..." );
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Optimize( 100 );
	_parent.SetStatusText( "Optimize complete; " + _parent.GetMeshHistory().
		GetCurrentMesh().GetModelStatisticsString() );*/
}
void DeformDialog::_cb_mirror()
{
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Mirror( _xaxis.Value(), _yaxis.Value(),
	                                                  _zaxis.Value() );
}
void DeformDialog::_cb_flip_normals()
{
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().FlipNormals();
}
void DeformDialog::_cb_unify()
{
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().Unify();
}
void DeformDialog::_cb_to_triangles()
{
	_parent.SetStatusText( "Converting quads to triangles..." );
	_parent.GetMeshHistory().AddMesh();
	_parent.GetMeshHistory().GetCurrentMesh().ToTriangles();
	_parent.SetStatusText( "Triangle conversion complete; " +
		_parent.GetMeshHistory().GetCurrentMesh().GetModelStatisticsString() );
}

void DeformDialog::_cb_pt()
{
	const unsigned its = 4;

	struct timeval beg, end;

	float dur = 0;

	for( unsigned i = 1; i <= its; i++ )
	{
		_parent.SetStatusText( "Running test " + ToString( (float)i ) );

		_parent.GetMeshHistory().AddMesh();
		gettimeofday( &beg, 0 );
		_parent.GetMeshHistory().GetCurrentMesh().PerformanceTest( "" );
		gettimeofday( &end, 0 );
		
		dur += ( end.tv_sec + 1.e-6 * end.tv_usec ) - 
			( beg.tv_sec + 1.e-6 * beg.tv_usec );
	}

	dur /= its;
	_parent.SetStatusText( "Tests complete; average run time: " + ToString( dur ) + "s." );

	_parent.View().Refresh();
	_parent.Refresh();
}
