/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "InputDialog.h"
#include "MainWindow.h"
#include "MessageDialog.h"
#include "Preferences.h"
#include "Repeater.h"
#include "Render.h"
#include "Utilities.h"
#include <GL/glu.h>
#include <sys/stat.h>

using namespace SharpConstruct::Interface;

MainWindow::MainWindow()
: TopWindow( 1024, 768, "SharpConstruct " + GetVersion() ),
  _mesh_history( 2 ),
  _v( 100, 100, 640, 480 ),
  _alpha_open_dialog( "Open brush alpha..." ),
  _deform_dialog( *this ),
  _new_dialog( *this ),
  _settings_dialog( *this ),
  _top_bar( DISABLED, 0, 4 ),
  
  _model( LARGEST_CHILD, 4 ),
  
  _brush_settings( DISABLED, 4, 4 ),
  _brush_settings_push( LARGEST_CHILD ),
 // _deform( 168, 200, "Deform" ),
  //_render( 168, 120, "Render" ),
  _view( LARGEST_CHILD, 4 ),
  
  _load( 64, "Open" ),
  _new( 64, "New" ),
  _save( 64, "Save" ),
  _save_as( "Save As" ),
  _undo( 64, "Undo" ),
  _redo( 64, "Redo" ),

  _displace( "Draw" ),
  _grab( "Grab" ),
  _smooth( "Smooth" ),
  _pinch( "Pinch" ),
  _deform( "Deform" ),
  _alpha( 80, 80, "" ),
  _radius( "Radius" ),
  _strength( "Strength" ),
  _intensity( "Intensity" ),
  _add( "Add" ),
  _sub( "Sub" ),
  _rgb( "RGB" ),
  _x_mirror( "X Symmetry" ),
  _y_mirror( "Y Symmetry" ),
  _z_mirror( "Z Symmetry" ),
  _color( 88 ),

  _wireframe( "Wireframe" ),
  _smoothed( "Smoothed" ),
  _double_sided( "Double Sided" ),
  _settings( "Settings" ),

  _rm_render( "Render" ),

  _about( "About" ),
  _model_info( "Info" ),
  _status_bar( "No model loaded - click 'New' or 'Open' to begin" ),
  
  _attempting_close( false )
{
	ResizeSignal().connect( sigc::mem_fun( _v, &ViewGL::Refresh ) );
	_mesh_history.Changed().connect( sigc::mem_fun( _v, &ViewGL::Refresh ) );
	//ResizeSignal.connect( sigc::mem_fun( *this, &MainWindow::_resize ) );

	_brush.SetPositionEdit( true );
	_initialize_widgets();

	_load_preferences();
	
	Mesh::MainBrush = &_brush;

	Resize();
}

MainWindow::~MainWindow()
{
	_save_preferences();
}

void MainWindow::_draw()
{
	glEnable( GL_ALPHA_TEST );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	glClearColor( 0.6, 0.6, 0.6, 1 );
	glClear( GL_COLOR_BUFFER_BIT );
	
	Reorganize();

	DrawChildren();
}

void MainWindow::Reorganize()
{
	glViewport( 0, 0, W(), H() );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( 0, Right(), Bottom(), 0 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef( -0.5, -0.5, 0 ); // Ah, that helps!
}

void MainWindow::Refresh()
{	
	_save.SetEnabled( !_mesh_history.GetCurrentMesh().Saved() );
	_save_as.SetEnabled( _v.Visible() );
	_undo.SetEnabled( _mesh_history.UndoPossible() );
	_redo.SetEnabled( _mesh_history.RedoPossible() );
	_deform.SetEnabled( _v.Visible() );
	_model_info.SetEnabled( _v.Visible() );
}

void MainWindow::RefreshAll()
{
	Refresh();
	_cb_update_brush_settings();
}

void MainWindow::Resize()
{
	TopWindow::Resize();

	_v.SetX( 4 );
	_v.SetY( 92 );
	_v.SetW( W() - 8 );
	_v.SetH( H() - 116 );
	
	_bottom_bar.SetY( H() - 20 );
	_bottom_bar.SetW( W() );
	_status_bar.AnchoredSetRight( Right() );
}

bool MainWindow::IsMeshDataSafe()
{
	std::string response;
	const bool SAFE = true;
	const bool UNSAFE = false;
	
	if( _mesh_history.GetCurrentMesh().Saved() )
		return SAFE;
	else
	{		
		response = MessageDialog::ModalMessage( *this, "Unsaved data",
			"The model is not saved.\nSave before continuing?",
			"Don't Save", "Save", "Cancel" );
		
		if( response == "Cancel" )
		{
			_status_bar.SetLabel( "Operation canceled" );
			return UNSAFE;
		}
		else if( response == "Save" )
		{
			_save_model();
			if( _mesh_history.GetCurrentMesh().Saved() )
				return SAFE;
			else
				return UNSAFE;				
		}
		else if( response == "Don't Save" )
			return SAFE;
		else
			return UNSAFE;
	}
}

SharpConstruct::MeshHistory& MainWindow::GetMeshHistory()
{
	return _mesh_history;
}

void MainWindow::SetModelPath( std::string p )
{
	_model_path = p;
}
ViewGL& MainWindow::View()
{
	return _v;
}
void MainWindow::SetStatusText( std::string t )
{
	_status_bar.SetLabel( t );
	Draw();
}
void MainWindow::DisplayModelInfo()
{
	_status_bar.SetLabel( _mesh_history.GetCurrentMesh().GetModelStatisticsString() );
}

void MainWindow::_initialize_view()
{
	_v.SetVisible( false );
	_v._history = &_mesh_history;
	_v.MainBrush = &_brush;
}

bool MainWindow::_open_model( std::string path )
{
	if( IsMeshDataSafe() )
	{
		SetModelPath( path );
		GetMeshHistory().Clear();
		GetMeshHistory().GetCurrentMesh().Load( path );
		View().Show();
		View().ResetView();
		RefreshAll();
		SetStatusText( FileDialog::FilePart( path ) + " loaded, " +
			GetMeshHistory().GetCurrentMesh().GetModelStatisticsString() );
		return true;
	}
	else
		_status_bar.SetLabel( "Open canceled" );
	
	return false;
}

void MainWindow::_save_model()
{
	if( _mesh_history.GetCurrentMesh().FileName().length() )
		_save_model_as( _mesh_history.GetCurrentMesh().FileName() );
	else
		_save_dialog.Show();
}

void MainWindow::_save_model_as( std::string path )
{
	if( path.length() )
	{
		_mesh_history.GetCurrentMesh().Save( path );
		_model_path = path;
		Refresh();
	}
}

void MainWindow::_initialize_widgets()
{
	_deform_dialog.SetVisible( false );
	_open_dialog.OpenSignal().connect( sigc::mem_fun( *this, &MainWindow::_open_model ) );
	_alpha_open_dialog.OpenSignal().connect( sigc::bind( sigc::mem_fun( _brush_editor, &BrushEditor::AddAlpha ), true ) );
	_new_dialog.SetVisible( false );
	_open_dialog.SetVisible( false );
	_open_dialog.ChangeDir( FileDialog::PathPart( _model_path ) );
	_alpha_open_dialog.SetVisible( false );
	_save_dialog.SaveSignal().connect( sigc::mem_fun( *this, &MainWindow::_save_model_as ) );
	_save_dialog.SetVisible( false );
	_settings_dialog.SetVisible( false );
	_dialogs.PushBack( _deform_dialog );
	_dialogs.PushBack( _new_dialog );
	_dialogs.PushBack( _open_dialog );
	_dialogs.PushBack( _alpha_open_dialog );
	_dialogs.PushBack( _save_dialog );
	_dialogs.PushBack( _settings_dialog );
	_v.ChangedMeshSignal.connect( sigc::mem_fun( *this, &MainWindow::Refresh ) );
	_brush_editor.ShowOpenDialogSignal().connect( sigc::mem_fun( _alpha_open_dialog, &OpenDialog::Show ) );
	_brush_editor.AlphaChangedSignal().connect( sigc::mem_fun( *this, &MainWindow::_update_brush_alpha ) );
	_update_brush_alpha( _brush_editor.CurrentAlpha() );
	
	_mesh_history.Changed().connect( sigc::mem_fun( *this, &MainWindow::Refresh ) );

	_model_left.PushBack( _load );
	_model_right.PushBack( _new );
	_model_left.PushBack( _save );
	_model_right.PushBack( _save_as );
	_model_left.PushBack( _undo );
	_model_right.PushBack( _redo );
	_model_left.PushBack( _model_info );
	_model.PushBack( _model_left );
	_model.PushBack( _model_right );
	
	_brush_settings_mode.PushBack( _displace );
	_brush_settings_mode.PushBack( _grab );
	_brush_settings_mode.PushBack( _smooth );
	_brush_settings_mode.PushBack( _pinch );
	_brush_settings_sliders.PushBack( _radius );
	_brush_settings_sliders.PushBack( _strength );
	_brush_settings_sliders.PushBack( _intensity );
	_brush_settings_push.PushBack( _add );
	_add.SetW( 67 );
	_brush_settings_push.PushBack( _sub );
	_brush_settings_push.PushBack( _rgb );
	_brush_settings_sliders.PushBack( _brush_settings_push );
	_brush_settings_symmetry.PushBack( _x_mirror );
	_brush_settings_symmetry.PushBack( _y_mirror );
	_brush_settings_symmetry.PushBack( _z_mirror );
	_brush_settings_symmetry.PushBack( _deform );
	_brush_settings.PushBack( _brush_settings_mode );
	_brush_settings.PushBack( _alpha );
	_brush_settings.PushBack( _brush_settings_sliders );
	_brush_settings.PushBack( _brush_settings_symmetry );
	
	_view.PushBack( _wireframe );
	_view.PushBack( _smoothed );
	_view.PushBack( _double_sided );
	_view.PushBack( _settings );

       	_external.PushBack( _rm_render );
	
	_top_bar.PushBack( _model );
	_top_bar.PushBack( _brush_settings );
	_top_bar.PushBack( _color );
	_top_bar.PushBack( _view );
	//_top_bar.PushBack( _external );
	
	_bottom_bar.PushBack( _about );
	_bottom_bar.PushBack( _status_bar );
	
	PushBack( _v );
	PushBack( _top_bar );
	PushBack( _bottom_bar );
	PushBack( GetRepeater() );
	PushBack( _brush_editor );
	PushBack( _dialogs );

	_load.SetShortcut( KeyData( 'o', false, true ) );
	_new.SetShortcut( KeyData( 'n', false, true ) );
	_save.SetShortcut( KeyData( 's', false, true ) );
	_save_as.SetShortcut( KeyData( 's', true, true ) );
	_undo.SetShortcut( KeyData( 'z', false, true ) );
	_redo.SetShortcut( KeyData( 'z', true, true ) );
	_displace.SetShortcut( KeyData( 'q' ) );
	_grab.SetShortcut( KeyData( 'w' ) );
	_smooth.SetShortcut( KeyData( 'e' ) );
	_pinch.SetShortcut( KeyData( 'r' ) );
	_deform.SetShortcut( KeyData( 'd', false, true ) );
	_radius.SetShortcut( KeyData( 'f' ) );
	_strength.SetShortcut( KeyData( 'd' ) );
	_intensity.SetShortcut( KeyData( 's' ) );
	_x_mirror.SetShortcut( KeyData( 'x' ) );
	_y_mirror.SetShortcut( KeyData( 'y' ) );
	_z_mirror.SetShortcut( KeyData( 'z' ) );
	_wireframe.SetShortcut( KeyData( 'w', false, true ) );

	_load.GetTooltip().SetMessage( "Open a model... <CTRL+O>",
				       "Supported formats include: OBJ" );
      	_new.GetTooltip().SetMessage( "Create a new model from one of several primitives... <CTRL+N>" );
	_save.GetTooltip().SetMessage( "Save the model under the current name. <CTRL+S>" );
	_save_as.GetTooltip().SetMessage( "Save the model under a new name... <CTRL+SHIFT+S",
					  "Supported formats include: OBJ, WRL" );
	_undo.GetTooltip().SetMessage( "Undo the last model operation. <CTRL+Z>",
				       "Note that the total number of undo operations can be set in the Settings Dialog." );
	_redo.GetTooltip().SetMessage( "Redo the last model operation. <CTRL+SHIFT+Z>" );
	_displace.GetTooltip().SetMessage( "Displace a part of the model. <Q>",
					   "This tool finds the average normal of all polygons beneath the brush, "
					   "an displaces those polygons along that normal, in amounts controled by "
					   "the brush alpha." );
	_grab.GetTooltip().SetMessage( "Grab a part of the model. <W>",
				       "This tool slides the polygons beneath the brush across the screen. "
				       "The polygons will not be moved any nearer or farther away, only "
				       "horizontally and vertically. Add and Sub work identically." );
	_smooth.GetTooltip().SetMessage( "Smooth a part of the model. <E>",
					 "Add and Sub work identically." );
	_pinch.GetTooltip().SetMessage( "Pinch a part of the model. <R>",
					"Add pinches polygons, but Sub expands them." );
	_deform.GetTooltip().SetMessage( "Apply a deformation to the entire model... <CTRL+D>" );
	_alpha.GetTooltip().SetMessage( "Select/load the shape of the brush..." );
	_radius.GetTooltip().SetMessage( "Set the size of the area affected by the brush. <F>" );
	_strength.GetTooltip().SetMessage( "Set the amount the brush affects the model's form. <D>" );
	_intensity.GetTooltip().SetMessage( "Set the amount the brush affects the model's color. <S>" );
	_add.GetTooltip().SetMessage( "When on, changes to the model's form occur in a positive direction. Toggled by <SHIFT>" );
	_sub.GetTooltip().SetMessage( "When on, changes to the model's form occur in a negative direction. Toggled by <SHIFT>" );
	_rgb.GetTooltip().SetMessage( "When on, brushes can affect colors." );
	_x_mirror.GetTooltip().SetMessage( "When on, changes to the model made with the brush will be duplicated over the X axis. <X>" );
	_y_mirror.GetTooltip().SetMessage( "When on, changes to the model made with the brush will be duplicated over the Y axis. <Y>" );
	_z_mirror.GetTooltip().SetMessage( "When on, changes to the model made with the brush will be duplicated over the Z axis. <Z>" );
	_wireframe.GetTooltip().SetMessage( "When on, the outlines of each polygon are superimposed over the model. <CTRL+W>" );
	_smoothed.GetTooltip().SetMessage( "When on, the edges of the polygons are blurred together." );
	_double_sided.GetTooltip().SetMessage( "When on, polygons can be viewed from either side, independent of their normals." );
	_settings.GetTooltip().SetMessage( "Set several preferences that control the way the programs looks and works." );
	_model_info.GetTooltip().SetMessage( "Displays model statistics in the status bar." );
	_about.GetTooltip().SetMessage( "Displays program information in the bar to the right." );
	
	// Load button
	_load.MouseUpSignal.connect( sigc::mem_fun( _open_dialog, &OpenDialog::Show ) );
	// New button
	_new.MouseUpSignal.connect( sigc::mem_fun( _new_dialog, &NewDialog::Show ) );
	// Save button
	_save.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_save_model ) );
	_save.SetEnabled( false );
	// Save As button
	_save_as.MouseUpSignal.connect( sigc::mem_fun( _save_dialog, &SaveDialog::Show ) );
	_save_as.SetEnabled( false );
	// Undo button
	_undo.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_undo ) );
	_undo.SetEnabled( false );
	// Redo button
	_redo.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_redo ) );
	_redo.SetEnabled( false );
	
	_displace.SetValue( 1 );
	_displace.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_draw ) );
	_grab.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_grab ) );
	_smooth.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_smooth ) );
	_pinch.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_pinch ) );
	_deform.SetEnabled( false );
	_deform.MouseUpSignal.connect( sigc::mem_fun( _deform_dialog, &DeformDialog::Show ) );
	_alpha.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_show_brush_editor ) );
	_radius.SetProperties( 1, 200, 1 );
	_radius.Value( 50 );
	_radius.ChangedSignal().connect( sigc::hide< -1 >( sigc::mem_fun( *this, &MainWindow::_cb_update_brush_settings ) ) );
	_strength.SetProperties( 1, 100, 1 );
	_strength.Value( 25 );
	_strength.ChangedSignal().connect( sigc::hide< -1 >( sigc::mem_fun( *this, &MainWindow::_cb_update_brush_settings ) ) );
	_intensity.SetProperties( 1, 100, 1 );
	_intensity.Value( 25 );
	_intensity.ChangedSignal().connect( sigc::hide< -1 >( sigc::mem_fun( *this, &MainWindow::_cb_update_brush_settings ) ) );
	_add.PressToggle();
	_add.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_add ) );
	_sub.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_sub ) );
	_rgb.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_rgb ) );
	_x_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_symmetry ) );
	_y_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_symmetry ) );
	_z_mirror.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_cb_symmetry ) );
	_color.ChangedSignal.connect( sigc::mem_fun( _brush, &Brush::SetColor ) );
	
	_wireframe.MouseUpSignal.connect( sigc::mem_fun( _v, &ViewGL::ToggleWireframe ) );
	_smoothed.MouseUpSignal.connect( sigc::mem_fun( _v, &ViewGL::ToggleSmoothed ) );
	_smoothed.SetValue( true );
	_double_sided.MouseUpSignal.connect( sigc::mem_fun( _v, &ViewGL::ToggleDoubleSided ) );
	_double_sided.SetValue( true );
	_settings.MouseUpSignal.connect( sigc::mem_fun( _settings_dialog, &SettingsDialog::Show ) );

	_rm_render.MouseUpSignal.connect( sigc::mem_fun( *this, &MainWindow::_render_handler ) );
	
	_model_info.MouseUpSignal.connect( sigc::mem_fun( this, &MainWindow::DisplayModelInfo ) );
	_model_info.SetEnabled( false );
	_about.MouseUpSignal.connect( sigc::mem_fun( this, &MainWindow::_display_about ) );
	
	_initialize_view();
	_cb_update_brush_settings();
}

void MainWindow::_display_about()
{
	_status_bar.SetLabel( "SharpConstruct v" + SharpConstruct::GetVersion() +
			      ", Copyright Nicholas Bishop, Licensed under the GPL -" +
			      "http://www.sharp3d.sourceforge.net" );
}

bool MainWindow::Close()
{
	if( _attempting_close )
		return false;
	
	_attempting_close = true;
	
	if( IsMeshDataSafe() )
	{
		_save_preferences();
		return true;
	}
	
	_attempting_close = false;
	
	return false;
}

void MainWindow::_cb_undo()
{
	_mesh_history.Undo();
	Refresh();
}

void MainWindow::_cb_redo()
{
	_mesh_history.Redo();
	Refresh();
}

void MainWindow::_cb_draw()
{
	_brush.SetModelingMode( SharpConstruct::Brush::DISPLACE );
	_displace.SetValue( 1 );
	_grab.SetValue( 0 );
	_smooth.SetValue( 0 );
	_pinch.SetValue( 0 );
}

void MainWindow::_cb_grab()
{
	_brush.SetModelingMode( SharpConstruct::Brush::GRAB );
	_displace.SetValue( 0 );
	_grab.SetValue( 1 );
	_smooth.SetValue( 0 );
	_pinch.SetValue( 0 );
}

void MainWindow::_cb_smooth()
{
	_brush.SetModelingMode( SharpConstruct::Brush::SMOOTH );
	_displace.SetValue( 0 );
	_grab.SetValue( 0 );
	_smooth.SetValue( 1 );
	_pinch.SetValue( 0 );
}

void MainWindow::_cb_pinch()
{
	_brush.SetModelingMode( SharpConstruct::Brush::PINCH );
	_displace.SetValue( 0 );
	_grab.SetValue( 0 );
	_smooth.SetValue( 0 );
	_pinch.SetValue( 1 );
}

void MainWindow::_cb_show_brush_editor()
{
	_brush_editor.SetX( _alpha.X() );
	_brush_editor.SetY( _alpha.Bottom() );
	_brush_editor.SetVisible( !_brush_editor.Visible() );
}

void MainWindow::_update_brush_alpha( Texture& t )
{
	_alpha.Icon().SetTexture( t );
	_brush.SetAlpha( t );
}

void MainWindow::_cb_update_brush_settings()
{
	float n_strength = _strength.Value() / -1000;
	
	if( !_add.Value() && !_sub.Value() )
		_brush.SetPositionEdit( false );
	else if( _sub.Value() )
	{
		n_strength = -n_strength;
		_brush.SetPositionEdit( true );
	}
	else if( _add.Value() )
	{
		_brush.SetPositionEdit( true );
	}
	
	_brush.SetRadius( _radius.Value() );
	_brush.SetStrength( n_strength );
	_brush.SetIntensity( _intensity.Value() / 100 );
	_v.RefreshCursor();
}

void MainWindow::_cb_add()
{
	if( _add.Value() && _sub.Value() )
		_sub.PressToggle();
	
	_cb_update_brush_settings();
}

void MainWindow::_cb_sub()
{
	if( _add.Value() && _sub.Value() )
		_add.PressToggle();
	
	_cb_update_brush_settings();
}

void MainWindow::_cb_rgb()
{
	_brush.SetColorEdit( _rgb.Value() );
}

void MainWindow::_cb_symmetry()
{
	_brush.SetSymmetryX( _x_mirror.Value() );
	_brush.SetSymmetryY( _y_mirror.Value() );
	_brush.SetSymmetryZ( _z_mirror.Value() );
}

void MainWindow::_render_handler()
{
	ExternalRender r;
	r.Render( _mesh_history.GetCurrentMesh() );
}

void MainWindow::_load_preferences()
{	
	SetW( Preferences::Instance().Get( "W", 1024 ) );
	SetH( Preferences::Instance().Get( "H", 768 ) );
	_model_path = Preferences::Instance().Get( "MODEL_PATH",
	                                           FileDialog::HomePath() );
	_open_dialog.ChangeDir( FileDialog::PathPart( _model_path ) );
	_save_dialog.ChangeDir( FileDialog::PathPart( _model_path ) );
	_alpha_open_dialog.ChangeDir( Preferences::Instance().Get( "ALPHA_PATH", FileDialog::HomePath() ) );

	_v.SetBackgroundColor( Color(
		Preferences::Instance().Get( "VIEW_COLOR_R", (float)0.3f ),
		Preferences::Instance().Get( "VIEW_COLOR_G", (float)0.3f ),
		Preferences::Instance().Get( "VIEW_COLOR_B", (float)0.3f ) ) );
	
	_mesh_history.SetMaximum( Preferences::Instance().Get( "HISTORY_STEPS", 2 ) );
}

void MainWindow::_save_preferences()
{
	Preferences::Instance().Set( "W", W() );
	Preferences::Instance().Set( "H", H() );
	Preferences::Instance().Set( "MODEL_PATH", _model_path );
	Preferences::Instance().Set( "ALPHA_PATH", _alpha_open_dialog.GetPath() );
}
