/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Mesh.h"
#include "Polygon3D.h"
#include "Utilities.h"
#include <cmath>
#include <fstream>
#include <sstream>
#include <string>

using namespace SharpConstruct;
using Optimized::Normal3D;
using Optimized::Point3D;

Normal3D Mesh::_view_up_normal( 0, 0, 0 );
Normal3D Mesh::_view_right_normal( 0, 0, 0 );
SharpConstruct::Brush *Mesh::MainBrush = 0;
Point3D Mesh::_current_selection( 0, 0, 0 );
//Magick::Image *Mesh::_texture = NULL;

Mesh::Mesh()
: _saved( true )
{}

/** Loads any supported mesh type.
 *
 * Parses the filename string to find the extension, then calls the appropriate
 * function to do the actual loading.
**/
void Mesh::Load( std::string filename )
{
	std::string extension = filename.substr( filename.rfind( '.',
	                                         std::string::npos ) + 1 );
	
	transform( extension.begin(), extension.end(), extension.begin(), toupper );

	if( extension == "OBJ" )
		_load_obj( filename );
	else
		_load_obj( filename );
	
	_filename = filename;
	_saved = true;
}

/** Saves any supported mesh type.
 *
 * Parses the filename string to find the extension, then calls the appropriate
 * function to do the actual saving.
**/
void Mesh::Save( std::string filename )
{
	std::string extension = filename.substr( filename.rfind( '.',
	                                         std::string::npos ) + 1 );
	
	transform( extension.begin(), extension.end(), extension.begin(), toupper );

	if( extension == "OBJ" )
		_save_obj( filename );
	else if( extension == "WRL" )
		_save_wrl( filename );
	else
		_save_obj( filename );
	
	_filename = filename;
	_saved = true;
}

void Mesh::BecomePlane( int columns, int rows )
{
	_filename = "";
	
	int i, j, t;
	const float half_x = columns / 2, half_y = rows / 2;
	
	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	//_vertex_uvs.clear();
	
	_clear_polys();
	
	_vertex_normals.resize( rows * columns );
	_vertex_colors.resize( rows * columns );
	for( i = 0; i < rows; i++ )
	{
		for( j = 0; j < columns; j++ )
		{
			_vertex_locations.push_back( Point3D( ( j - half_x ) / ( columns - 1 ), ( i - half_y ) / ( rows - 1 ), 0 ) );

			//_vertex_uvs.push_back( UV( 1.0 * j / ( rows - 1 ), 1 - ( 1.0 * i / ( rows - 1 ) ) ) );
		}
	}

	_quad_normals.resize( ( rows - 1 ) * ( columns - 1 ) );
	
	for( i = 0, t = 0; i < rows - 1; i++ )
	{
		for( j = 0; j < columns - 1; j++, t++ )
		{
			_quads.push_back( Quad( i * columns + j + 1,
			                        ( i + 1 ) * columns + j + 1,
			                        ( i + 1 ) * columns + j,
			                        i * columns + j ) );
		}
	}
		
	_recalculate_vertex_users();
	RecalculateAllNormals();

	Unify();
	Flood( 1 );

	/*struct timeval tp1, tp2;
	gettimeofday( &tp1, 0 );
	for( i = 0; i < 30000; i++ )
		_recalculate_all_normals();
	gettimeofday( &tp2, 0 );
	std::cout << ( tp2.tv_sec + 1.e-6 * tp2.tv_usec ) - 
	             ( tp1.tv_sec + 1.e-6 * tp1.tv_usec ) << std::endl;*/
	/*_recalculate_all_normals();

	Unify();
	Flood( 1 );*/

	_change_operation();
}

void Mesh::BecomeSphere( unsigned subdivisions )
{
	_filename = "";
	
	_clear_polys();
	
	// Creates an octahedron
	_vertex_locations.resize( 6 );
	_vertex_normals.resize( 6 );
	_vertex_colors.resize( 6 );
	
	_triangle_normals.resize( 8 );
	
	_vertex_locations[ 0 ] = Point3D( 0, 1, 0 );
	_vertex_locations[ 1 ] = Point3D( 1, 0, 0 );
	_vertex_locations[ 2 ] = Point3D( 0, -1, 0 );
	_vertex_locations[ 3 ] = Point3D( -1, 0, 0 );
	_vertex_locations[ 4 ] = Point3D( 0, 0, 1 );
	_vertex_locations[ 5 ] = Point3D( 0, 0, -1 );
	
	_triangles.push_back( Triangle( 4, 1, 0 ) );
	_triangles.push_back( Triangle( 4, 2, 1 ) );
	_triangles.push_back( Triangle( 4, 3, 2 ) );
	_triangles.push_back( Triangle( 4, 0, 3 ) );
	_triangles.push_back( Triangle( 5, 0, 1 ) );
	_triangles.push_back( Triangle( 5, 1, 2 ) );
	_triangles.push_back( Triangle( 5, 2, 3 ) );
	_triangles.push_back( Triangle( 5, 3, 0 ) );

	_recalculate_vertex_users();
	
	for( unsigned i = 1; i < subdivisions; i++ )
		Subdivide( false );
	
	// Normalizing the points makes a sphere with a radius of 1
	for( unsigned i = 0; i < _vertex_locations.size(); i++ )
		_vertex_locations[ i ].Normalize();
	
	RecalculateAllNormals();
	
	Flood( 1 );
	
	_change_operation();
}

void Mesh::BecomeCube( unsigned subdivisions )
{
	_filename = "";
	
	// Creates an eight-vertex cube
	_clear_polys();
	_vertex_locations.clear();
	_vertex_normals.resize( 8 );
	_vertex_colors.resize( 8 );
	_quad_normals.resize( 6 );
	
	_vertex_locations.push_back( Point3D( 1, 1, -1 ) );
	_vertex_locations.push_back( Point3D( 1, -1, -1 ) );
	_vertex_locations.push_back( Point3D( -1, -1, -1 ) );
	_vertex_locations.push_back( Point3D( -1, 1, -1 ) );
	_vertex_locations.push_back( Point3D( 1, 1, 1 ) );
	_vertex_locations.push_back( Point3D( 1, -1, 1 ) );
	_vertex_locations.push_back( Point3D( -1, -1, 1 ) );
	_vertex_locations.push_back( Point3D( -1, 1, 1 ) );
	
	_quads.push_back( Quad( 0, 1, 2, 3 ) );
	_quads.push_back( Quad( 4, 7, 6, 5 ) );
	_quads.push_back( Quad( 0, 4, 5, 1 ) );
	_quads.push_back( Quad( 1, 5, 6, 2 ) );
	_quads.push_back( Quad( 2, 6, 7, 3 ) );
	_quads.push_back( Quad( 4, 0, 3, 7 ) );
	
	_recalculate_vertex_users();
	
	for( unsigned i = 1; i < subdivisions; i++ )
		Subdivide( false );
	
	RecalculateAllNormals();
	
	Flood( 1 );
	_change_operation();
}

/*void Mesh::SetTexture( Magick::Image *t )
{
	_texture = t;
}*/

sigc::signal< void >& Mesh::Changed()
{
	return _changed;
}
void Mesh::_change_operation()
{
	_saved = false;
	Changed()();
}
std::string Mesh::FileName() const
{
	return _filename;
}
bool Mesh::Saved() const
{
	return _saved;
}

/** Returns mesh data information.
 *
 * Returns a string in this format:
 * "[number of vertices] vert[ex/ices], [number of polygons] polygon[s]"
**/
std::string Mesh::GetModelStatisticsString()
{
	std::ostringstream oss;

	oss << _vertex_locations.size();
	if( _vertex_locations.size() == 1 )
		oss << " vertex, ";
	else
		oss << " vertices, ";
	oss << _poly_count();
	if( _poly_count() == 1 )
		oss << " polygon";
	else
		oss << " polygons";

	return oss.str();
}

float Mesh::GetDiameter()
{
	_calculate_bounding_sphere();

	return _diameter;
}

void Mesh::CopyEssentials( Mesh &mesh )
{
	_vertex_locations = mesh._vertex_locations;
	_vertex_normals = mesh._vertex_normals;
	_vertex_colors = mesh._vertex_colors;
	//_vertex_uvs = mesh._vertex_uvs;
	_triangle_normals = mesh._triangle_normals;
	_quad_normals = mesh._quad_normals;
	_triangles = mesh._triangles;
	_quads = mesh._quads;
	_vertex_users = mesh._vertex_users;
	_saved = mesh._saved;
	_filename = mesh._filename;
	_txtr = mesh._txtr;
	_selection_radius = mesh._selection_radius;
}

/** Free unimportant mesh data.
 *
 * Free memory by clearing temporary data unimportant to a mesh object.
**/
void Mesh::ClearExtraData()
{
	_active_vertex_indices.clear();
	_active_vertex_distances.clear();
	_modified_vertex_indices.clear();
}

void Mesh::PerformanceTest( std::string )
{
	for( unsigned i = 0; i < 200; i++ )
	{
		_current_selection = _vertex_locations[ rand() % _vertex_locations.size() ];
		DisplaceAreas();
		_find_active_vertices( _active_vertex_indices, _active_vertex_distances,
	                               _get_mirror( false, false, false ) );
	}
	RecalculateAllNormals();

	_change_operation();
}

/** Loads a mesh in the OBJ format.
 *
 * Loads mesh data from an OBJ file. Support is incomplete; the parsing has
 * bugs, and normal data is ignored.
**/
void Mesh::_load_obj( std::string filename )
{
	std::ifstream file( filename.c_str(), std::ios::in );
	//std::istringstream line;
	std::string token, subtoken;
	char buffer[ 256 ];
	unsigned vertex_count = 0, polygon_count = 0, uvs_count;
	unsigned find_loc;
	unsigned ndx;
	bool hascolor = false;

	_vertex_locations.clear();
	_vertex_normals.clear();
	_vertex_colors.clear();
	_uvs.clear();
	_clear_polys();
	std::vector< Polygon3D > polygons;

	while( !file.eof() )
	{
		std::istringstream line;
		file.getline( buffer, 256 );
		line.str( buffer );
		line >> token;
		if( token == "v" )
		{
			float x, y, z;
			line >> x >> y >> z;
			_vertex_locations.push_back( Point3D( x, y, z ) );
		}
		else if( token == "vt" )
		{
			float u, v;
			line >> u >> v;
			_uvs.push_back( UV( u, v ) );
		}
		else if( token == "f" )
		{
			polygons.resize( polygon_count + 1 );
			while( line >> token )
			{
				int loc = token.find( '/' );
				subtoken = token.substr( 0, loc );
				polygons[ polygon_count ].VertexIndices.push_back( ToInt( subtoken ) - 1 );

				//std::cout << token << " " << token.length() << " " << loc << std::endl;
				
				/*if( loc == std::string::npos )
					continue;

				token = token.substr( loc + 1 );

				std::cout << token << std::endl;*/
				/*
				loc = token.find( '/' ) + 1;
				subtoken = token.substr( 0, loc );
				polygons[ polygon_count ].UVIndices.push_back( ToInt( subtoken ) - 1 );*/
			}
			/*for( unsigned z = 0; z < polygons[ polygon_count ].VertexIndices.size(); z++ )
				std::cout << polygons[ polygon_count ].VertexIndices[ z ] << " ";
				std::cout << std::endl;*/
			polygon_count++;
		}
		else if( token == "#vertexcolor" )
		{
			line >> ndx;
			_vertex_colors.push_back( SharpConstruct::Color() );
			line >> _vertex_colors[ ndx ].Red >> _vertex_colors[ ndx ].Green >>
			        _vertex_colors[ ndx ].Blue;
			hascolor = true;
		}
	}
	file.close();

	/*if( u.size() && v.size() )
	{
		for( unsigned i = 0; i < u.size(); i++ )
		{
			_vertices.U[ i ] = u[ i ];
			_vertices.V[ i ] = v[ i ];
		}
	}*/
	
	_vertex_normals.resize( _vertex_locations.size() );
	
	for( unsigned i = 0; i < polygons.size(); i++ )
	{
		if( polygons[ i ].VertexIndices.size() == 3 )
		{
			_triangles.push_back( Triangle( polygons[ i ].VertexIndices[ 0 ],
			                                polygons[ i ].VertexIndices[ 1 ],
			                                polygons[ i ].VertexIndices[ 2 ] ) );
		}
		else if( polygons[ i ].VertexIndices.size() == 4 )
		{
			_quads.push_back( Quad( polygons[ i ].VertexIndices[ 0 ],
			                        polygons[ i ].VertexIndices[ 1 ],
			                        polygons[ i ].VertexIndices[ 2 ],
			                        polygons[ i ].VertexIndices[ 3 ] ) );
		}
		else
		{
			for( unsigned j = 1; j < polygons[ i ].VertexIndices.size() - 1; j++ )
			{
				_triangles.push_back( Triangle( polygons[ i ].VertexIndices[ 0 ],
				                                polygons[ i ].VertexIndices[ j ],
				                                polygons[ i ].VertexIndices[ j + 1 ] ) );
			}
		}
	}
	
	_triangle_normals.resize( _triangles.size() );
	_quad_normals.resize( _quads.size() );
	
	_recalculate_vertex_users();
	RecalculateAllNormals();
	
	if( !hascolor )
	{
		_vertex_colors.resize( _vertex_locations.size() );
		Flood( 1 );
	}
	
	_saved = true;
}

/** Saves a mesh in the OBJ format.
 *
 * Saves mesh data to an OBJ file. Only vertex and polygon data is exported,
 * not normals. The top line is a comment including the SharpConstruct URL.
**/
void Mesh::_save_obj( std::string filename )
{
	std::ofstream file( filename.c_str(), std::ios::out );
	
	int i, j;
	
	file << "# Exported from SharpConstruct " + SharpConstruct::GetVersion() +
	        " - http://sharp3d.sourceforge.net/\n\n";
	
	for( i = 0; i < _vertex_locations.size(); i++ )
		file << "v " << _vertex_locations[ i ].X() << ' ' <<
	                    _vertex_locations[ i ].Y() << ' ' <<
	                    _vertex_locations[ i ].Z() << '\n';
	
	//for( i = 0; i < _vertex_locations.size(); i++ )
	//	file << "vt " << _vertices.U[ i ] << ' ' << _vertices.V[ i ] << '\n';
	
	for( i = 0; i < _triangles.size(); i++ )
	{
		file << "f ";
		for( j = 0; j <= 2; j++ )
			file << _triangles[ i ].V[ j ] + 1 << ' ';		
		file << '\n';
	}
	for( i = 0; i < _quads.size(); i++ )
	{
		file << "f ";
		for( j = 0; j <= 3; j++ )
			file << _quads[ i ].V[ j ] + 1 << ' ';		
		file << '\n';
	}
	
	for( i = 0; i < _vertex_colors.size(); i++ )
		file << "#vertexcolor " << i << ' ' << _vertex_colors[ i ].Red << ' '
		     << _vertex_colors[ i ].Green << ' ' << _vertex_colors[ i ].Blue
		     << '\n';
	
	file.close();
	
	_saved = true;
}

/** Saves a mesh in the VRML97 (.wrl) format.
 *
 * Saves mesh data to an WRL file. Only vertex and polygon data is exported,
 * not normals. The top line is the VRML magic numbers string,
 * followed by a line with a comment including the SharpConstruct URL.
 *
 * Thanks to Joerg Scheurich for this code.
**/
void Mesh::_save_wrl( std::string filename )
{
	std::ofstream file( filename.c_str(), std::ios::out );
	
	int i, j;
	
	file << "#VRML V[ 1 ].0 utf8\n# Exported from SharpConstruct " + SharpConstruct::GetVersion() +
	        " - http://sharp3d.sourceforge.net/\n\n";
	file << "Transform\n";
	file << "  {\n";
	file << "  children\n";
	file << "    [\n";
	file << "    Shape\n";
	file << "      {\n";
	file << "      appearance       Appearance\n";
	file << "        {\n";
	file << "        material        Material\n";
	file << "          {\n";
	file << "          }\n";
	file << "        }\n";
	file << "      geometry       IndexedFaceSet\n";
	file << "        {\n";
	file << "        # double sided\n";
	file << "        solid FALSE\n";
	file << "        # single sided: inside or outside ?\n";
	file << "        #ccw TRUE\n";
	file << "        #ccw FALSE\n";
	file << "        # smooth if angle < 90 degree\n";
	file << "        creaseAngle 1.57\n";
	file << "        coord   Coordinate\n";
	file << "          {\n";
	file << "          point\n";
	file << "            [\n";
	for( i = 0; i < _vertex_locations.size(); ++i )
	{
		file << "            " << 
		_vertex_locations[ i ].X() << ", " <<
		_vertex_locations[ i ].Y() << ", " <<
		_vertex_locations[ i ].Z() << ", \n";
	}
	file << "            ]\n";
	file << "          }\n";
	file << "        coordIndex\n";
	file << "          [\n";
	for( i = 0; i < _triangles.size(); ++i )
	{
		file << "          ";
		for( j = 0; j <= 2; j++ )
			file << _triangles[ i ].V[ j ] << ", ";		
		file << " -1, " << '\n';
	}
	for( i = 0; i < _quads.size(); ++i )
	{
		file << "          ";
		for( j = 0; j <= 3; j++ )
			file << _quads[ i ].V[ j ] << ", ";		
		file << " -1, " << '\n';
	}
	file << "          ]\n";

/* not implemented yet	
for( i = 0; i < _vertices.size(); i++ )
	file << "#vertexcolor " << i << ' ' << _vertices[ i ].R << ' ' <<
			_vertices[ i ].G << ' ' << _vertices[ i ].B << '\n';
*/
	file << "        }\n";
	file << "      }\n";
	file << "    ]\n";
	file << "  }\n";

	file.close();
	
	_saved = true;
}

Point3D Mesh::_vertex_neighbor_loc_average( unsigned v )
{
	Point3D average( 0, 0, 0 );
	unsigned total = 0;
	for( unsigned j = 0; j < _vertex_users[ v ].Triangles.size(); j++ )
	{
		const unsigned polygon_index = _vertex_users[ v ].Triangles[ j ];
		for( unsigned k = 0; k <= 2; k++ )
			if( _triangles[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 2;
				if( next_point_index == 3 )
					next_point_index = 0;
				
				average += _vertex_locations[ v ];
				average += _vertex_locations[ _triangles[ polygon_index ].V[ previous_point_index ] ];
				average += _vertex_locations[ _triangles[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	for( unsigned j = 0; j < _vertex_users[ v ].Quads.size(); j++ )
	{
		const unsigned polygon_index = _vertex_users[ v ].Quads[ j ];
		for( unsigned k = 0; k <= 3; k++ )
			if( _quads[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 3;
				if( next_point_index == 4 )
					next_point_index = 0;
				
				average += _vertex_locations[ v ];
				average += _vertex_locations[ _quads[ polygon_index ].V[ previous_point_index ] ];
				average += _vertex_locations[ _quads[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	
	average /= total;
	
	return average;
}
Color Mesh::_vertex_neighbor_clr_average( unsigned v )
{
	Color average( 0, 0, 0 );
	unsigned total = 0;
	for( unsigned j = 0; j < _vertex_users[ v ].Triangles.size(); j++ )
	{
		const unsigned polygon_index = _vertex_users[ v ].Triangles[ j ];
		for( unsigned k = 0; k <= 2; k++ )
			if( _triangles[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 2;
				if( next_point_index == 3 )
					next_point_index = 0;
				
				average += _vertex_colors[ v ];
				average += _vertex_colors[ _triangles[ polygon_index ].V[ previous_point_index ] ];
				average += _vertex_colors[ _triangles[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	for( unsigned j = 0; j < _vertex_users[ v ].Quads.size(); j++ )
	{
		const unsigned polygon_index = _vertex_users[ v ].Quads[ j ];
		for( unsigned k = 0; k <= 3; k++ )
			if( _quads[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 3;
				if( next_point_index == 4 )
					next_point_index = 0;
				
				average += _vertex_colors[ v ];
				average += _vertex_colors[ _quads[ polygon_index ].V[ previous_point_index ] ];
				average += _vertex_colors[ _quads[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	
	average /= total;
	
	return average;
}

Point3D Mesh::_polygon_center( Triangle& t )
{
	Point3D c( 0, 0, 0 );
	for( unsigned i = 0; i <= 2; ++i )
		c += _vertex_locations[ t.V[ i ] ];
	c /= 3;
	return c;
}
Point3D Mesh::_polygon_center( Quad& q )
{
	Point3D c( 0, 0, 0 );
	for( unsigned i = 0; i <= 3; ++i )
		c += _vertex_locations[ q.V[ i ] ];
	c /= 4;
	return c;
}

/** Calculates center and diameter of sphere formed by vertices.
 *
 * All vertices in the model are used to create a virtual sphere, with a center
 * at the average of all the points and the diameter equal to two times the
 * distance from the center to the furthest point from the center.
**/
void Mesh::_calculate_bounding_sphere()
{
	int i;
	float tmp;

	_diameter = 0;
	
	_center.Zero();

	for( i = 0; i < _vertex_locations.size(); i++ )
		_center += _vertex_locations[ i ];
	
	_center /= _vertex_locations.size();
	
	for( i = 0; i < _vertex_locations.size(); i++ )
	{
		tmp = _center.Distance( _vertex_locations[ i ] );
		if( tmp > _diameter )
			_diameter = tmp;
	}
	
	_diameter *= 2;
}

/** Calculates the average normal of the area.
 *
 * Returns the average normal of all the points within the area.
**/
Point3D Mesh::_calculate_area_normal()
{
	Normal3D area_normal;
	area_normal.Zero();
		
	for( unsigned i = 0; i < _active_vertex_indices.size(); i++ )
		area_normal += _vertex_normals[ _active_vertex_indices[ i ] ];
	
	area_normal.Normalize();
	
	return area_normal;
}

/** Updates all vertex and polygon normals.
 *
 * Acts like _recalculate_modified_normals(), but does the recalculation for
 * every vertex and every polygon.
 *
 * Note: all polygons are treated as planar
**/
void Mesh::RecalculateAllNormals()
{
	for( int i = 0; i < _triangle_normals.size(); i++ )
		_triangle_normals[ i ].CalculatePlaneUnormal(
			_vertex_locations[ _triangles[ i ].V[ 0 ] ].RawData(),
			_vertex_locations[ _triangles[ i ].V[ 1 ] ].RawData(),
			_vertex_locations[ _triangles[ i ].V[ 2 ] ].RawData() );
	_triangle_normals.NormalizeAll();
	for( int i = 0; i < _quad_normals.size(); i++ )
		_quad_normals[ i ].CalculatePlaneUnormal(
			_vertex_locations[ _quads[ i ].V[ 0 ] ].RawData(),
			_vertex_locations[ _quads[ i ].V[ 1 ] ].RawData(),
			_vertex_locations[ _quads[ i ].V[ 2 ] ].RawData() );
	_quad_normals.NormalizeAll();
	unsigned j, k;
	for( unsigned i = 0; i < _vertex_users.size(); i++ )
	{
		_vertex_normals[ i ].Zero();
		for( j = 0; j < _vertex_users[ i ].Triangles.size(); j++ )
			_vertex_normals[ i ] += _triangle_normals[ _vertex_users[ i ].Triangles[ j ] ];
		for( k = 0; k < _vertex_users[ i ].Quads.size(); k++ )
			_vertex_normals[ i ] += _quad_normals[ _vertex_users[ i ].Quads[ k ] ];

		_vertex_normals[ i ] /= j + k;
	}
}

void Mesh::_recalculate_damaged()
{
	for( unsigned i = 0; i < _modified_vertex_indices.size(); i++ )
	{
		for( unsigned j = 0; j < _vertex_users[ _modified_vertex_indices[ i ] ].Triangles.size(); j++ )
		{
			_damaged_triangles.push_back( _vertex_users[ _modified_vertex_indices[ i ] ].Triangles[ j ] );
		}
		for( unsigned j = 0; j < _vertex_users[_modified_vertex_indices[ i ] ].Quads.size(); j++ )
		{
			_damaged_quads.push_back( _vertex_users[ _modified_vertex_indices[ i ] ].Quads[ j ] );
		}
	}
	_damaged_triangles.sort();
	_damaged_triangles.unique();
	_damaged_quads.sort();
	_damaged_quads.unique();
}

/** Updates modified vertex and polygon normals.
 *
 * Uses the _modified_vertex_indices list to update only modified vertex and
 * polygon normals. This is generally much faster than calling
 * RecalculateAllNormals().
**/
void Mesh::_recalculate_modified_normals()
{
	unsigned i, j;
	
	if( _modified_vertex_indices.size() < _vertex_locations.size() )
	{
		for( std::list< unsigned >::iterator a = _damaged_triangles.begin();
		     a != _damaged_triangles.end(); a++ )
		{
			_triangle_normals[ *a ].CalculatePlaneNormal(
				_vertex_locations[ _triangles[ ( *a ) ].V[ 0 ] ].RawData(),
				_vertex_locations[ _triangles[ ( *a ) ].V[ 1 ] ].RawData(),
				_vertex_locations[ _triangles[ ( *a ) ].V[ 2 ] ].RawData() );
		}
		for( std::list< unsigned >::iterator a = _damaged_quads.begin();
		     a != _damaged_quads.end(); a++ )
		{
			_quad_normals[ *a ].CalculatePlaneNormal(
				_vertex_locations[ _quads[ ( *a ) ].V[ 0 ] ].RawData(),
				_vertex_locations[ _quads[ ( *a ) ].V[ 1 ] ].RawData(),
				_vertex_locations[ _quads[ ( *a ) ].V[ 2 ] ].RawData() );
		}
		for( i = 0; i < _modified_vertex_indices.size(); i++ )
		{
			unsigned count = 0;
			_vertex_normals[ _modified_vertex_indices[ i ] ].Zero();
			for( j = 0; j < _vertex_users[ _modified_vertex_indices[ i ] ].Triangles.size(); j++ )
			{
				count++;
				_vertex_normals[ _modified_vertex_indices[ i ] ]+=
					_triangle_normals[ _vertex_users[ _modified_vertex_indices[ i ] ].Triangles[ j ] ];
			}
			for( j = 0; j < _vertex_users[ _modified_vertex_indices[ i ] ].Quads.size(); j++ )
			{
				count++;
				_vertex_normals[ _modified_vertex_indices[ i ] ]+=
					_quad_normals[ _vertex_users[ _modified_vertex_indices[ i ] ].Quads[ j ] ];
			}
			if( count > 0 )
				_vertex_normals[ _modified_vertex_indices[ i ] ] /= count;
		}
	}
	else
		RecalculateAllNormals();
}

void Mesh::_recalculate_vertex_users()
{
	_vertex_users.clear();
	_vertex_users.resize( _vertex_locations.size() );
	for( unsigned i = 0; i < _triangles.size(); i++ )
	{
		_vertex_users[ _triangles[ i ].V[ 0 ] ].Triangles.push_back( i );
		_vertex_users[ _triangles[ i ].V[ 1 ] ].Triangles.push_back( i );
		_vertex_users[ _triangles[ i ].V[ 2 ] ].Triangles.push_back( i );
	}
	for( unsigned i = 0; i < _quads.size(); i++ )
	{
		_vertex_users[ _quads[ i ].V[ 0 ] ].Quads.push_back( i );
		_vertex_users[ _quads[ i ].V[ 1 ] ].Quads.push_back( i );
		_vertex_users[ _quads[ i ].V[ 2 ] ].Quads.push_back( i );
		_vertex_users[ _quads[ i ].V[ 3 ] ].Quads.push_back( i );
	}
}

std::map< unsigned, unsigned > Mesh::_triangle_users( unsigned v1, unsigned v2 )
{
	std::map< unsigned, unsigned > r;
	// For every user of v1
	for( unsigned i = 0; i < _vertex_users[ v1 ].Triangles.size(); i++ )
	{
		// For every user of v2
		for( unsigned j = 0; j < _vertex_users[ v2 ].Triangles.size(); j++ )
		{
			const unsigned poly = _vertex_users[ v1 ].Triangles[ i ];
			// If poly is a user of both v2 (as well as v1)
			if( poly == _vertex_users[ v2 ].Triangles[ j ] )
			{
				unsigned w;
				for( unsigned k = 0; k < 3; k++ )
				{
					// Return the side (1, 2, or 3) that's connected
					
					if( _triangles[ poly ].V[ k ] != v1 &&
					    _triangles[ poly ].V[ k ] != v2 )
					{
						w = k + 2;
						if( k == 2 )
							w = 1;
						break;
					}
				}
				r[ poly] = w;
				break;
			}
		}
	}
	
	return r;
}
std::map< unsigned, unsigned > Mesh::_quad_users( unsigned v1, unsigned v2 )
{
	std::map< unsigned, unsigned > r;
	// For every user of v1
	for( unsigned i = 0; i < _vertex_users[ v1 ].Quads.size(); i++ )
	{
		// For every user of v2
		for( unsigned j = 0; j < _vertex_users[ v2 ].Quads.size(); j++ )
		{
			const unsigned poly = _vertex_users[ v1 ].Quads[ i ];
			// If poly is a user of both v2 (as well as v1)
			if( poly == _vertex_users[ v2 ].Quads[ j ] )
			{
				unsigned w;
				for( unsigned k = 0; k < 4; k++ )
				{
					unsigned k2 = k + 1;
					if( k2 > 3 )
						k2 = 0;
					if( ( _quads[ poly ].V[ k ] == v1 &&
					      _quads[ poly ].V[ k2 ] == v2 ) ||
					    ( _quads[ poly ].V[ k ] == v2 &&
					      _quads[ poly ].V[ k2 ] == v1 ) )
					{
						w = k + 1;
						break;
					}
				}
				r[ poly] = w;
				break;
			}
		}
	}
	
	return r;
}

unsigned Mesh::_poly_count() const
{
	return _triangles.size() + _quads.size();
}

void Mesh::_clear_polys()
{
	_triangles.clear();
	_quads.clear();
	_triangle_normals.clear();
	_quad_normals.clear();
}

std::set< unsigned > Mesh::_connected_vertices( unsigned v )
{
	/**
	std::set< unsigned > r;
	
	//std::cout << _vertex_polygon_indices[ v ].size() << ": ";
	for( unsigned i = 0; i < _vertex_polygon_indices[ v ].size(); ++i )
	{
		const unsigned polygon_index = _vertex_polygon_indices[ v ][ i ];
		for( unsigned j = 0; j <= 2; ++j )
		{
			if( _indices[ polygon_index * 3 + j ] != v )
			{
				r.insert( _indices[ polygon_index * 3 + j ] );
			}
		}
	}
	//std::cout << r.size() << std::endl;
	return r;**/
}

/** Stores vertices within the MainBrush->GetRadius().
 *
 * Stores a list of all vertices within the sphere that has its center at area
 * and a radius of MainBrush->GetRadius(). The distance from the active vertices
 * to the center is also stored.
**/
// Slow
void Mesh::_find_active_vertices( std::vector< unsigned > &vertex_indices,
                                  std::vector< float > &vertex_distances,
                                  EditData e )
{
	vertex_indices.clear();
	vertex_distances.clear();
	
	float distance;
	
	for( int i = 0; i < _vertex_locations.size(); i++ )
	{
		distance = e.Center.Distance( _vertex_locations[ i ] );
		if( distance < _selection_radius )
		{
			vertex_indices.push_back( i );
			vertex_distances.push_back( distance );
		}
	}
	
	_selected_vertex[ e.SIndex ] = vertex_indices[ 0 ];

	_modified_vertex_indices.insert( _modified_vertex_indices.end(),
					 vertex_indices.begin(),
	                                 vertex_indices.end() );
}
/* This version of _find_active_vertices may be rather slow too...*/
/*void Mesh::_find_active_vertices( std::vector< unsigned > &vertex_indices,
                                  std::vector< float > &vertex_distances,
                                  EditData e )
{	
	vertex_indices.clear();
	vertex_distances.clear();
	
	int area_index = 0;

	if( _selected_vertex[ e.SIndex ] >= _vertex_locations.size() )
		_selected_vertex[ e.SIndex ] = 0;

	// If a point used last time through the loop is within the Radius, use it!
	if( _vertex_locations[ _selected_vertex[ e.SIndex ] ].Distance( e.Center ) <
	    _selection_radius )
	{
		area_index = _selected_vertex[ e.SIndex ];
	}
	else
	{
		float tmp_distance;
		
		// The point of this loop is just to get any vertex that is within the Radius
		for( int i = 0; i < _vertex_locations.size(); i++ )
		{
			e.Center.FastDistance( _vertex_locations[ i ], tmp_distance );

			if( tmp_distance < _selection_radius )
			{
				area_index = i;
				break;
			}
		}
	}

	std::set< unsigned > checked;
	ActiveVertexData d( area_index, checked, vertex_indices, vertex_distances, e.Center );
	_is_vertex_active( d );
	
	_selected_vertex[ e.SIndex ] = vertex_indices[ 0 ];

	_modified_vertex_indices.insert( _modified_vertex_indices.end(),
					 vertex_indices.begin(),
	                                 vertex_indices.end() );
}*/

void Mesh::_is_vertex_active( ActiveVertexData& d )
{
	unsigned current = d.Current;

	int xdf = d.Checked.size();
	d.Checked.insert( current );
	
	float distance = d.Area.Distance( _vertex_locations[ current ] );
	if( distance < _selection_radius )
	{
		if( current < _lowest_index )
			_lowest_index = current;
		if( current > _highest_index )
			_highest_index = current;
		d.VertexIndices.push_back( current );
		d.VertexDistances.push_back( distance );
	}
	else
		return;
	
	int j, polygon_index;
	int previous_point_index;
	// Loop through each triangle using the current point
	for( unsigned i = 0; i < _vertex_users[ current ].Triangles.size(); i++ )
	{
		// Set the current polygon
		polygon_index =_vertex_users[ current ].Triangles[ i ];
		// Loop through each vertex in the polygon
		for( j = 0; j <= 2; j++ )
			// If the current active point equals the current VertexIndex...
			if( _triangles[ polygon_index ].V[ j ] == current )
			{
				previous_point_index = j - 1;
				if( previous_point_index < 0 )
					previous_point_index = 2;

				previous_point_index = _triangles[ polygon_index ].V[ previous_point_index ];

				if( d.Checked.find( previous_point_index ) == d.Checked.end() )
				{
					d.Current = previous_point_index;
					_is_vertex_active( d );
				}
				break;
			}
	}
	
	// Loop through each quad using the current point
	for( unsigned i = 0; i < _vertex_users[ current ].Quads.size(); i++ )
	{
		// Set the current polygon
		polygon_index =_vertex_users[ current ].Quads[ i ];
		// Loop through each vertex in the polygon
		for( j = 0; j <= 3; j++ )
			// If the current active point equals the current VertexIndex...
			if( _quads[ polygon_index ].V[ j ] == current )
			{
				previous_point_index = j - 1;
				if( previous_point_index < 0 )
					previous_point_index = 3;

				previous_point_index = _quads[ polygon_index ].V[ previous_point_index ];

				if( d.Checked.find( previous_point_index ) == d.Checked.end() )
				{
					d.Current = previous_point_index;
					_is_vertex_active( d );
				}
				break;
			}
	}
}
