/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Texture.h"
#include <GL/gl.h>
#include <X11/Xlib.h>
#include <Imlib2.h>
#include <cmath>
#include <fstream>

using namespace SharpConstruct;

int SharpConstruct::POTSize( int npot )
{
	float pot = pow( 2, round( log( npot ) / log( 2 ) ) );
	if( pot < npot )
		return pot * 2;
	else
		return pot;
}

Texture::Texture()
: _handle( 0 ), _glbuffer( 0 )
{}

Texture::~Texture()
{
	if( _handle )
	{
		imlib_context_set_image( _handle );
		imlib_free_image();
	}
	
	if( _glbuffer )
		glDeleteTextures( 1, &_glbuffer );
}
			
bool Texture::Load( std::string in )
{
	_path = in;

	_handle = imlib_load_image_without_cache( in.c_str() );
	if( !_handle )
		return false;

       imlib_context_set_image( _handle );

	if( _glbuffer )
		glDeleteTextures( 1, &_glbuffer );

	glGenTextures( 1, &_glbuffer );
	glBindTexture( GL_TEXTURE_2D, _glbuffer );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	glBindTexture( GL_TEXTURE_2D, _glbuffer );
	glTexImage2D( GL_TEXTURE_2D, 0, 4, imlib_image_get_width(),
				  imlib_image_get_height(), 0, GL_BGRA, GL_UNSIGNED_BYTE,
				  imlib_image_get_data_for_reading_only() );

	return true;
}
void Texture::AutoCorrect()
{
	int sz = W();
	if( H() > W() )
		sz = H();
	sz = POTSize( sz + 1 ) / 2;
	if( W() == H() && sz == W() )
		return;
	imlib_context_set_image( _handle );
	int org_w = W(), org_h = H();
	Imlib_Image cpy = imlib_create_image( sz, sz );
	imlib_context_set_image( cpy );
	imlib_blend_image_onto_image( _handle, 0, 0, 0, org_w, org_h, 0, 0, sz, sz );
	imlib_context_set_image( _handle );
	imlib_free_image();
	_handle = cpy;
	imlib_context_set_image( _handle );

	glBindTexture( GL_TEXTURE_2D, _glbuffer );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	glBindTexture( GL_TEXTURE_2D, _glbuffer );
	glTexImage2D( GL_TEXTURE_2D, 0, 4, imlib_image_get_width(),
				  imlib_image_get_height(), 0, GL_BGRA, GL_UNSIGNED_BYTE,
				  imlib_image_get_data_for_reading_only() );
}

void Texture::Bind() const
{
	if( _handle )
	{
		imlib_context_set_image( _handle );
		
		glBindTexture( GL_TEXTURE_2D, _glbuffer );
	}
}

unsigned int* Texture::RawData()
{
	imlib_context_set_image( _handle );
	
	return imlib_image_get_data();
}
unsigned* Texture::ReadOnlyRawData()
{
	imlib_context_set_image( _handle );
	
	return imlib_image_get_data_for_reading_only();
}
void Texture::PutDataBack( unsigned int* d )
{
	imlib_image_put_back_data( d );
}

unsigned Texture::W() const
{
	if( _handle )
	{
		imlib_context_set_image( _handle );
		return imlib_image_get_width();
	}
	else
		return 0;
}
unsigned Texture::H() const
{
	if( _handle )
	{
		imlib_context_set_image( _handle );
		return imlib_image_get_height();
	}
	else
		return 0;
}

std::string Texture::Path() const
{
	return _path;
}
