/* Copyright 2004, 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "ArcBall.h"
#include "MeshHistory.h"
#include "MessageDialog.h"
#include "Preferences.h"
#include "TopWindow.h"
#include "View.h"
#include "OglExt.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <cmath>

using SharpConstruct::Interface::ViewGL;

using SharpConstruct::Optimized::OPT_SIZE;
using SharpConstruct::Optimized::Point3D;
using SharpConstruct::Optimized::Normal3D;

ViewGL::ViewGL( int x, int y, int w, int h )
: Widget( x, y, w, h, "" ), _rendered( false ), _store_view( true ),
  _display_copied_view( true ), _last_mouse_location( 0, 0 ),
  _recalc_projected( true ), _only_recalc_damaged( false ),
  _first_run( true ), _initialized( false ),
  _vertex_buffer( 0 ), _triangle_buffer( 0 ), _quad_buffer( 0 ), _use_vbo( true ),
  _update_vertex_buffer( true ), _update_index_buffer( true ), _arc_ball( 0.8 ),
  _background_color( 0.3, 0.3, 0.3 ), _smoothed( true ), _double_sided( true )
{
	SetCursor( CROSS );
	MouseDownSignal.connect( sigc::mem_fun( *this, &ViewGL::_mouse_down ) );
	MouseUpSignal.connect( sigc::mem_fun( *this, &ViewGL::_mouse_up ) );
	MouseMoveSignal.connect( sigc::mem_fun( *this, &ViewGL::_mouse_move ) );
	ScrollDownSignal.connect( sigc::mem_fun( *this, &ViewGL::_scroll_down ) );
	ScrollUpSignal.connect( sigc::mem_fun( *this, &ViewGL::_scroll_up ) );
}

ViewGL::~ViewGL()
{
	if( _vertex_buffer )
		glDeleteBuffersARB( 1, &_vertex_buffer );
	if( _triangle_buffer )
		glDeleteBuffersARB( 1, &_triangle_buffer );
	if( _quad_buffer )
		glDeleteBuffersARB( 1, &_quad_buffer );
}

void ViewGL::KeyDown()
{
	if( Event::Keyboard.Key == ' ' )
	{
		ResetView();
	}
}

void ViewGL::ResetView()
{
	Mesh& mesh = _history->GetCurrentMesh();

	mesh._calculate_bounding_sphere();
	_location = mesh._center;
	_cam_distance = mesh._diameter;
	_arc_ball.Reset();
	_update_vertex_buffer = true;
	_recalc_projected = true;
	_rendered = false;
	_update_index_buffer = true;
}

void ViewGL::Refresh()
{
	_initialized = false;
	_update_vertex_buffer = true;
	_recalc_projected = true;
	_rendered = false;
	_update_index_buffer = true;
	_history->GetCurrentMesh()._modified_vertex_indices.clear();
}

void ViewGL::RefreshCursor()
{
	_prepare_cursor();
}

void ViewGL::SetBackgroundColor( Color input )
{
	_background_color = input;
}

const SharpConstruct::Color& ViewGL::BackgroundColor() const
{
	return _background_color;
}

void ViewGL::ToggleWireframe()
{
	_wireframe = !_wireframe;
	Refresh();
}

void ViewGL::ToggleSmoothed()
{
	_smoothed = !_smoothed;
	Refresh();
}

void ViewGL::ToggleDoubleSided()
{
	_double_sided = !_double_sided;
	Refresh();
}

void ViewGL::_draw_interface()
{
	_initialize();
	glClearColor( _background_color.Red, _background_color.Green,
	              _background_color.Blue, 1 );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	_orientate_view();

	if( _wireframe )
	{
		glEnable( GL_POLYGON_OFFSET_FILL );
		glPolygonOffset( 1, 1 );
		_draw_mesh();
		glDisable( GL_POLYGON_OFFSET_FILL );
	
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
		glBlendFunc( GL_ZERO, GL_ZERO );
		
		_draw_mesh();
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
		
	}
	else
		_draw_mesh();
}

void ViewGL::_draw()
{
	//glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, _framebuffer );

	Mesh& mesh = _history->GetCurrentMesh();
	
	_make_view_context();

	//if( !_rendered || mesh._damaged_triangles.size() > 0 ||
	//    mesh._damaged_quads.size() > 0 )
	{
		//std::cout << "FullRender" << std::endl;
		//glClear( GL_ACCUM_BUFFER_BIT );
		//glDrawBuffer( GL_AUX1 );

		mesh._damaged_triangles.clear();
		mesh._damaged_quads.clear();

		_draw_interface();
		_find_projected_points();
		_find_view_normals();

		_rendered = true;
	}
	
	/* Debug code, draws bounding box around changed polygons
	if( _damaged_area.size() >= 4 )
	{		
		glColor3f( 1, 0, 0 );
		glBegin( GL_LINE_STRIP );
		{
			glVertex2i( _damaged_area[ 0 ] + 4, Display().H() - 24 - _damaged_area[ 1 ] );
			glVertex2i( _damaged_area[ 2 ] + 4, Display().H() - 24 - _damaged_area[ 1 ] );
			glVertex2i( _damaged_area[ 2 ] + 4, Display().H() - 24 - _damaged_area[ 3 ] );
			glVertex2i( _damaged_area[ 0 ] + 4, Display().H() - 24 - _damaged_area[ 3 ] );
			glVertex2i( _damaged_area[ 0 ] + 4, Display().H() - 24 - _damaged_area[ 1 ] );
		}
		glEnd();
	}*/
	
	/*Mesh& mesh = _history->GetCurrentMesh();
	if( mesh._damaged_polygon_indices.size() > 0 )
	{
		_only_recalc_damaged = true;
		_recalc_projected = true;
		_make_view_context();
		_orientate_view();
		_find_projected_points();
		_revert_view_context();
		
		
		glDrawBuffer( GL_AUX1 );
		
				glScissor( _damaged_area[ 0 ],
		           _damaged_area[ 1 ],
		           _damaged_area[ 2 ] - _damaged_area[ 0 ],
		           _damaged_area[ 3 ] - _damaged_area[ 1 ] );
		glEnable( GL_SCISSOR_TEST );
		//glClear( GL_DEPTH_BUFFER_BIT );
		
		// Clear depth buffer in damaged area that we're about to redraw...
		glViewport( 0, 0, Display().W(), Display().H() );
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		gluOrtho2D( Display().X(), Display().Right(),
					Display().Y(), Display().Bottom() );
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		glEnable( GL_DEPTH_TEST );
		glDepthFunc( GL_ALWAYS );
		glBegin( GL_TRIANGLES );
		glColor3f( 0, 0, 0 );
		for( std::list< unsigned >::iterator i = mesh._damaged_polygon_indices.begin();
		     i != mesh._damaged_polygon_indices.end(); i++ )
		{
			for( unsigned j = 0; j < 3; j++ )
			{
				glVertex3f( _projected_points[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							_projected_points[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							-1 );
			}
		}
		glEnd();

		_make_view_context();
		_orientate_view();
		glDepthFunc( GL_LESS );			
		
		// Find the full (view-based) damage...
		std::list< unsigned > damaged_polygon_indices;
		for( unsigned i = 0; i < _projected_points.size(); i++ )
		{
			Point3D& p = _projected_points[ i ];
			if( p.X() >= _damaged_area[ 0 ] &&
			    p.Y() >= _damaged_area[ 1 ] &&
			    p.X() <= _damaged_area[ 2 ] &&
			    p.Y() <= _damaged_area[ 3 ] )
			{
				for( unsigned j = 0; j < mesh._vertex_polygon_indices[ i ].size(); j++ )
					damaged_polygon_indices.push_back( mesh._vertex_polygon_indices[ i ][ j ] );
			}
		}
		damaged_polygon_indices.unique();
		
		Mesh& mesh = _history->GetCurrentMesh();

		
		glBegin( GL_TRIANGLES );
		for( std::list< unsigned >::iterator i = damaged_polygon_indices.begin();
		     i != damaged_polygon_indices.end(); i++ )
		{
			for( unsigned j = 0; j < 3; j++ )
			{
				glNormal3f( mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			
				glColor3f( mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Red,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Green,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Blue );
			
				glVertex3f( mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			}
		}
		glEnd();
		//glDisable( GL_SCISSOR_TEST );
		glBegin( GL_TRIANGLES );
		for( std::list< unsigned >::iterator i = mesh._damaged_polygon_indices.begin();
		     i != mesh._damaged_polygon_indices.end(); i++ )
		{
			for( unsigned j = 0; j < 3; j++ )
			{
				glNormal3f( mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_normals[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			
				glColor3f( mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Red,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Green,
						   mesh._vertex_colors[ mesh._indices[ ( *i ) * 3 + j ] ].Blue );
			
				glVertex3f( mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].X(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Y(),
							mesh._vertex_locations[ mesh._indices[ ( *i ) * 3 + j ] ].Z() );
			}
		}
		glEnd();
		
		mesh._damaged_polygon_indices.clear();
		_find_projected_points();
	}*/

	/*if( mesh._damaged_triangles.size() > 0 )
	{
		std::cout << "PartialRender" << std::endl;
		glDrawBuffer( GL_AUX1 );
		//glDrawBuffer( GL_BACK );
		//glReadBuffer( GL_BACK );
		_make_view_context();
		_orientate_view();
		//glDepthMask( GL_FALSE );
		//glDisable( GL_DEPTH_TEST );
		glBegin( GL_TRIANGLES );
		for( std::list< unsigned >::iterator i = mesh._damaged_triangles.begin();
		     i != mesh._damaged_triangles.end(); i++ )
		{
			for( unsigned j = 0; j < 3; j++ )
			{
				glNormal3f( mesh._vertex_normals[ mesh._triangles[ *i ].V[ j ] ].X(),
					    mesh._vertex_normals[ mesh._triangles[ *i ].V[ j ] ].Y(),
					    mesh._vertex_normals[ mesh._triangles[ *i ].V[ j ] ].Z() );
			
				glColor3f( mesh._vertex_colors[ mesh._triangles[ *i ].V[ j ] ].Red,
					   mesh._vertex_colors[ mesh._triangles[ *i ].V[ j ] ].Green,
					   mesh._vertex_colors[ mesh._triangles[ *i ].V[ j ] ].Blue );
			
				glVertex3f( mesh._vertex_locations[ mesh._triangles[ *i ].V[ j ] ].X(),
					    mesh._vertex_locations[ mesh._triangles[ *i ].V[ j ] ].Y(),
					    mesh._vertex_locations[ mesh._triangles[ *i ].V[ j ] ].Z() );
			}
		}
		glEnd();
		_find_projected_points();
		mesh._damaged_triangles.clear();
		//glDepthMask( GL_TRUE );
		}*/

	/*glDrawBuffer( GL_BACK );
	glReadBuffer( GL_AUX1 );
	_revert_view_context();
	glRasterPos2f( X(), Y() + H() );
	glCopyPixels( 0, 0, W(), H(), GL_COLOR );

	glDrawBuffer( GL_BACK );
	glReadBuffer( GL_BACK );
	glDisable( GL_SCISSOR_TEST );*/

	// Draw circular-cursor boundary
	_revert_view_context();
       	if( Contains( Event::Mouse.Location ) )
		glTranslatef( Event::Mouse.Location.X(), Event::Mouse.Location.Y(), 0 );
	else
		glTranslatef( X() + W() / 2, Y() + H() / 2, 0 );
	glCallList( _cursor_list );
	_revert_view_context();

	// Draw symmetry-mode markers
	glBegin( GL_TRIANGLES );
	{
		for( unsigned i = 0; i < _symm_locs.size(); i++ )
		{
			const Math::Point2D< float > sloc( X() + _symm_locs[ i ].X(), TopWindow::Display().H() - _symm_locs[ i ].Y() );
			glColor3f( 0, 0, 0 );
			glVertex2f( sloc.X(), sloc.Y() + 3 );
			glColor3f( 1, 1, 1 );
			glVertex2f( sloc.X() + 3, sloc.Y() - 3 );
			glColor3f( 1, 0, 0 );
			glVertex2f( sloc.X() - 3, sloc.Y() - 3 );
		} glEnd();
	}
	
	/*glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

	glBindTexture( GL_TEXTURE_2D, _copy_tex );
	//mesh._txtr.Bind();

	glEnable( GL_TEXTURE_2D );
	glColor3f( 1, 1, 1 );
	glBegin( GL_QUADS );
	{
		glTexCoord2f( 0, 1 );
		glVertex2f( X(), Y() );

		glTexCoord2f( 1, 1 );
		glVertex2f( X() + W(), Y() );

		glTexCoord2f( 1, 0 );
		glVertex2f( X() + W(), Y() + H() );

		glTexCoord2f( 0, 0 );
		glVertex2f( X(), Y() + H() );
	} glEnd();
	glDisable( GL_TEXTURE_2D );*/

	//glFinish();
}

void ViewGL::_make_view_context()
{
	glEnable( GL_COLOR_MATERIAL );
	glEnable( GL_LIGHTING );
	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LESS );
	
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glViewport( X(), TopWindow::Display().H() - Bottom(), W(), H() );
	float ratio = W();
	if( H() )
		ratio /= H();
	gluPerspective( 45, ratio, 0.01, 100 );

	glEnable( GL_SCISSOR_TEST );
	glScissor( X(), TopWindow::Display().H() - Bottom(), W(), H() );
}
void ViewGL::_revert_view_context()
{
	glViewport( 0, 0, TopWindow::Display().W(), TopWindow::Display().H() );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( TopWindow::Display().X(), TopWindow::Display().Right(),
	            TopWindow::Display().Bottom(), TopWindow::Display().Y() );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef( -0.5, -0.5, 0 );

	glDisable( GL_SCISSOR_TEST );
	glDisable( GL_CULL_FACE );
	glShadeModel( GL_SMOOTH );
	glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glDisable( GL_COLOR_MATERIAL );
}
void ViewGL::_orientate_view()
{
	double trackball_matrix[ 16 ];
	unsigned col, row, index;

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0, 0, ( _cam_distance - _location.Z() ) * 1.5, -_location.X(),
	           -_location.Y(), _location.Z(), 0, 1, 0 );

	const ArcBall::Transform* rot;
	if( Event::Keyboard.ShiftDown && Event::Mouse.RightButtonDown )
		rot = &_arc_ball.SnapRotation( _arc_ball.Transformation(), 45 );
	else
		rot = &_arc_ball.Transformation();

	for( col = 0, index = 0; col < 4; col++ )
	{
		for( row = 0; row < 4; row++ )
			trackball_matrix[ index++ ] = ( *rot )[ row ][ col ];
	}
  
	glMultMatrixd( trackball_matrix );
}

void ViewGL::_initialize()
{
	Mesh& mesh = _history->GetCurrentMesh();
	
	if( _first_run )
	{
		_first_run = false;

		if( !glexExtensionsSupported( "GL_ARB_vertex_buffer_object" ) )
		{
			_use_vbo = false;

			if( Preferences::Instance().Get( "ShowVBOError", std::string( "Yes" ) ) == "Yes" )
			{
				std::string answr;
				answr = MessageDialog::ModalMessage( TopWindow::Display(), "Graphics warning",
								     "Support for the OpenGL extension GL_ARB_vertex_buffer_object not found.\n"
								     "Rendering will use vertex arrays, which may or may not be slower. If\n"
								     "you believe your card should support this feature, try updating your\n"
								     "drivers.\n\nKeep displaying this message?", "Yes", "No" );
				Preferences::Instance().Set( "ShowVBOError",  answr );				
			}
		}

		GLfloat light_ambient[] = { 0.15, 0.15, 0.15, 1 };
		GLfloat light_diffuse[] = { 0.6, 0.6, 0.6, 1 };	
		GLfloat light_position_one[] = { 0, 0, -150 + _location.Z(), 0 };
		glLightfv( GL_LIGHT1, GL_AMBIENT, light_ambient );
		glLightfv( GL_LIGHT1, GL_DIFFUSE, light_diffuse );
		glLightfv( GL_LIGHT1, GL_POSITION, light_position_one );
		glEnable( GL_LIGHT1 );

		//glEnableClientState( GL_TEXTURE_COORD_ARRAY );
		glEnableClientState( GL_VERTEX_ARRAY );
		glEnableClientState( GL_NORMAL_ARRAY );
		glEnableClientState( GL_COLOR_ARRAY );

		_cursor_list = glGenLists( 1 );
	}
	
	// Useful for testing mesh functions, but should not be enabled for release:
	//glPolygonMode( GL_FRONT_AND_BACK, GL_POINT );

	RefreshCursor();
	
	if( _smoothed )
		glShadeModel( GL_SMOOTH );
	else
		glShadeModel( GL_FLAT );	
	
	glGetDoublev( GL_PROJECTION_MATRIX, _projection_matrix );
	glGetIntegerv( GL_VIEWPORT, _viewport );

	if( _double_sided )
	{
		glDisable( GL_CULL_FACE );
		glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
	}
	else
	{
		glEnable( GL_CULL_FACE );
		glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE );
	}

	if( _use_vbo )
	{
		if( _vertex_buffer == 0 )
			glGenBuffersARB( 1, &_vertex_buffer );
		glBindBufferARB( GL_ARRAY_BUFFER_ARB, _vertex_buffer );
		if( _update_vertex_buffer || _update_index_buffer )
		{
			_update_vertex_buffer = false;
			if( mesh._modified_vertex_indices.size() == 0 )
			{
				glBufferDataARB( GL_ARRAY_BUFFER_ARB,
						 mesh._vertex_locations.size() * OPT_SIZE * 2 +
						 mesh._vertex_colors.size() * sizeof( Color ),// +
						 //mesh._vertex_uvs.size() * sizeof( UV ),
						 0, GL_STREAM_DRAW_ARB );
				// Load vertex locations...
				glBufferSubDataARB( GL_ARRAY_BUFFER_ARB, 0,
						    mesh._vertex_locations.size() * OPT_SIZE,
						    mesh._vertex_locations.RawData() );
				glVertexPointer( 3, GL_FLOAT, OPT_SIZE, 0 );
				// Load vertex normals...
				glBufferSubDataARB( GL_ARRAY_BUFFER_ARB,
						    mesh._vertex_locations.size() * OPT_SIZE,
						    mesh._vertex_normals.size() * OPT_SIZE,
						    mesh._vertex_normals.RawData() );
				glNormalPointer( GL_FLOAT, OPT_SIZE,
						 ( ( char* )NULL +
						   ( mesh._vertex_locations.size() * OPT_SIZE ) ) );
				// Load vertex colors...
				glBufferSubDataARB( GL_ARRAY_BUFFER_ARB,
						    mesh._vertex_locations.size() * OPT_SIZE * 2,
						    mesh._vertex_colors.size() * sizeof( Color ),
						    &mesh._vertex_colors[ 0 ] );
				glColorPointer( 3, GL_FLOAT, 0,
						( ( char* )NULL +
						  ( mesh._vertex_locations.size() * OPT_SIZE * 2 ) ) );
				
				// Load vertex UV's...
				/*glBufferSubDataARB( GL_ARRAY_BUFFER_ARB,
						    mesh._vertex_locations.size() * OPT_SIZE * 2 +
						    mesh._vertex_colors.size() * sizeof( Color ),
						    mesh._vertex_uvs.size() * sizeof( UV ),
						    &mesh._vertex_uvs[ 0 ] );
				glTexCoordPointer( 2, GL_FLOAT, 0,
						   ( ( char* )NULL +
						   ( mesh._vertex_locations.size() * OPT_SIZE * 2 + mesh._vertex_colors.size() * sizeof( Color ) ) ) );*/
				
				//std::cout << "UpAll" << std::endl;
			}
			else
			{
				//std::cout << " partial" << std::endl;
				//std::cout << mesh._modified_vertex_indices.size() << std::endl;
				
				float* p = (float*)glMapBufferARB( GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY );
				if( MainBrush->GetPositionEdit() )
				{
					for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); i++ )
					{
						const unsigned ndx = mesh._modified_vertex_indices[ i ];
						mesh._vertex_locations[ ndx ].CopyTo( &p[ ndx * 4 ] );
					}
					const unsigned s = mesh._vertex_locations.size() * 4;
					for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); i++ )
					{
						const unsigned ndx = mesh._modified_vertex_indices[ i ];
						mesh._vertex_normals[ ndx ].CopyTo( &p[ s + ndx * 4 ] );
					}
				}
				if( MainBrush->GetColorEdit() )
				{
					const unsigned s = mesh._vertex_locations.size() * 8;
					for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); i++ )
					{
						const unsigned ndx = mesh._modified_vertex_indices[ i ];
						p[ s + ndx * 3 ] = mesh._vertex_colors[ ndx ].Red;
						p[ s + ndx * 3 + 1 ] = mesh._vertex_colors[ ndx ].Green;
						p[ s + ndx * 3 + 2 ] = mesh._vertex_colors[ ndx ].Blue;
					}
				}
				glUnmapBufferARB( GL_ARRAY_BUFFER_ARB );
			}			
		}
		if( _triangle_buffer == 0 )
			glGenBuffersARB( 1, &_triangle_buffer );
		if( _quad_buffer == 0 )
			glGenBuffersARB( 1, &_quad_buffer );
		if( _update_index_buffer )
		{
			glBindBufferARB( GL_ELEMENT_ARRAY_BUFFER_ARB, _triangle_buffer );
			glBufferDataARB( GL_ELEMENT_ARRAY_BUFFER_ARB, mesh._triangles.size() *
					 sizeof( Triangle ), &mesh._triangles[ 0 ],
					 GL_STATIC_DRAW_ARB );
			glBindBufferARB( GL_ELEMENT_ARRAY_BUFFER_ARB, _quad_buffer );
			glBufferDataARB( GL_ELEMENT_ARRAY_BUFFER_ARB, mesh._quads.size() *
					 sizeof( Quad ), &mesh._quads[ 0 ],
					 GL_STATIC_DRAW_ARB );
			_update_index_buffer = false;
		}
	}
	else // If we can't use VBOs, use simple array pointers
	{
		glVertexPointer( 3, GL_FLOAT, OPT_SIZE, mesh._vertex_locations.RawData() );
		glNormalPointer( GL_FLOAT, OPT_SIZE, mesh._vertex_normals.RawData() );
		glColorPointer( 3, GL_FLOAT, 0, &mesh._vertex_colors[ 0 ] );
		//glTexCoordPointer( 2, GL_FLOAT, 0, &mesh._vertex_uvs[ 0 ] );
	}
	
	_initialized = true;
}

void ViewGL::_prepare_cursor()
{
	const unsigned SEGMENTS = 32;
	glNewList( _cursor_list, GL_COMPILE );
	{
		glColor4f( 1, 0, 0, 0.5 );
		glBegin( GL_LINE_LOOP );
		{
			for( unsigned i = 0; i < SEGMENTS; ++i )
			{
				glVertex2f( MainBrush->GetRadius() * cos( i * 2 * M_PI / SEGMENTS ),
					    MainBrush->GetRadius() * sin( i * 2 * M_PI / SEGMENTS ) );	
			}
		} glEnd();
	} glEndList();
}

void ViewGL::_draw_mesh()
{
	using namespace SharpConstruct::Optimized;
	
	Mesh& mesh = _history->GetCurrentMesh();

	//glBindTexture( GL_TEXTURE_2D, _texture_buffer );
	
	/*std::cout << "Total: " << mesh._vertex_locations.size()
	          << ", Damaged: " << mesh._modified_vertex_indices.size()
	          << ", Damaged Range: " << mesh._highest_index - mesh._lowest_index << std::endl;*/

	//glEnable( GL_TEXTURE_2D );

	// Displays all the verts, no polygons involved
	/*glBegin( GL_TRIANGLES );
	for( unsigned i = 0; i < mesh._vertex_locations.size(); i++ )
	{
		glColor3f( mesh._vertex_colors[ i ].Red,
			   mesh._vertex_colors[ i ].Green,
			   mesh._vertex_colors[ i ].Blue );
		glNormal3f( mesh._vertex_normals[ i ].X(),
			    mesh._vertex_normals[ i ].Y(),
			    mesh._vertex_normals[ i ].Z() );
		glVertex3f( mesh._vertex_locations[ i ].X(),
			    mesh._vertex_locations[ i ].Y(),
			    mesh._vertex_locations[ i ].Z() );
	} glEnd();
	return;*/

	//mesh._txtr.Bind();
	if(_use_vbo )
	{
		glBindBufferARB( GL_ELEMENT_ARRAY_BUFFER_ARB, _triangle_buffer );
		glDrawRangeElements( GL_TRIANGLES, 0, mesh._triangles.size() * 3 - 1,
				     mesh._triangles.size() * 3, GL_UNSIGNED_INT, 0 );
		glBindBufferARB( GL_ELEMENT_ARRAY_BUFFER_ARB, _quad_buffer );
		glDrawRangeElements( GL_QUADS, 0, mesh._quads.size() * 4 - 1,
				     mesh._quads.size() * 4, GL_UNSIGNED_INT, 0 );
	}
	else // If we can't use VBO, use vertex arrays
	{
		glDrawElements( GL_TRIANGLES, mesh._triangles.size() * 3, GL_UNSIGNED_INT, &mesh._triangles[ 0 ] );
		glDrawElements( GL_QUADS, mesh._quads.size() * 4, GL_UNSIGNED_INT, &mesh._quads[ 0 ] );
	}
	if( 0 ) // Immediate mode (texture rendering?)
	{
		glBegin( GL_TRIANGLES );
		for( unsigned i = 0; i < mesh._triangles.size(); i++ )
		{
			for( unsigned j = 0; j <= 2; j++ )
			{
				glColor3f( mesh._vertex_colors[ mesh._triangles[ i ].V[ j ] ].Red,
					   mesh._vertex_colors[ mesh._triangles[ i ].V[ j ] ].Green,
					   mesh._vertex_colors[ mesh._triangles[ i ].V[ j ] ].Blue );
				glNormal3f( mesh._vertex_normals[ mesh._triangles[ i ].V[ j ] ].X(),
					    mesh._vertex_normals[ mesh._triangles[ i ].V[ j ] ].Y(),
					    mesh._vertex_normals[ mesh._triangles[ i ].V[ j ] ].Z() );
				//glTexCoord2f( mesh._vertex_uvs[ mesh._triangles[ i ].V[ j ] ].U,
				//	      mesh._vertex_uvs[ mesh._triangles[ i ].V[ j ] ].V );
				glVertex3f( mesh._vertex_locations[ mesh._triangles[ i ].V[ j ] ].X(),
					    mesh._vertex_locations[ mesh._triangles[ i ].V[ j ] ].Y(),
					    mesh._vertex_locations[ mesh._triangles[ i ].V[ j ] ].Z() );
			}
		} glEnd();
		glBegin( GL_QUADS );
		for( unsigned i = 0; i < mesh._quads.size(); i++ )
		{
			for( unsigned j = 0; j <= 3; j++ )
			{
				glColor3f( mesh._vertex_colors[ mesh._quads[ i ].V[ j ] ].Red,
					   mesh._vertex_colors[ mesh._quads[ i ].V[ j ] ].Green,
					   mesh._vertex_colors[ mesh._quads[ i ].V[ j ] ].Blue );
				glNormal3f( mesh._vertex_normals[ mesh._quads[ i ].V[ j ] ].X(),
					    mesh._vertex_normals[ mesh._quads[ i ].V[ j ] ].Y(),
					    mesh._vertex_normals[ mesh._quads[ i ].V[ j ] ].Z() );
				//glTexCoord2f( mesh._vertex_uvs[ mesh._quads[ i ].V[ j ] ].U,
				//	      mesh._vertex_uvs[ mesh._quads[ i ].V[ j ] ].V );
				glVertex3f( mesh._vertex_locations[ mesh._quads[ i ].V[ j ] ].X(),
					    mesh._vertex_locations[ mesh._quads[ i ].V[ j ] ].Y(),
					    mesh._vertex_locations[ mesh._quads[ i ].V[ j ] ].Z() );
			}
		} glEnd();
	}
	
	//glDisable( GL_TEXTURE_2D );
}

void ViewGL::_find_view_normals()
{
	GLdouble model_matrix[ 16 ];
	double projected_x, projected_y, projected_z;
	double zero_loc[ 3 ];
	
	glGetDoublev( GL_MODELVIEW_MATRIX, model_matrix );
	
	gluUnProject( 0, 0, 0, model_matrix, _projection_matrix, _viewport,
		        &zero_loc[ 0 ], &zero_loc[ 1 ], &zero_loc[ 2 ] );

	// Update Mesh::_view_right_normal
	gluUnProject( 1, 0, 0, model_matrix, _projection_matrix, _viewport,
	            &projected_x, &projected_y, &projected_z );
	Mesh::_view_right_normal = Normal3D( projected_x, projected_y, projected_z );
	Mesh::_view_right_normal.X() -= zero_loc[ 0 ];
	Mesh::_view_right_normal.Y() -= zero_loc[ 1 ];
	Mesh::_view_right_normal.Z() -= zero_loc[ 2 ];
	Mesh::_view_right_normal.Normalize();
	
	// Update Mesh::_view_up_normal
	gluUnProject( 0, -1, 0, model_matrix, _projection_matrix, _viewport,
	            &projected_x, &projected_y, &projected_z );
	Mesh::_view_up_normal = Normal3D( projected_x, projected_y, projected_z );
	Mesh::_view_up_normal.X() -= zero_loc[ 0 ];
	Mesh::_view_up_normal.Y() -= zero_loc[ 1 ];
	Mesh::_view_up_normal.Z() -= zero_loc[ 2 ];
	Mesh::_view_up_normal.Normalize();
}

void ViewGL::_find_projected_points()
{
	GLdouble model_matrix[ 16 ];
	double projected_x, projected_y, projected_z;
	Mesh& mesh = _history->GetCurrentMesh();

	glGetDoublev( GL_MODELVIEW_MATRIX, model_matrix );
	
	float depth;

	const Math::Point2D< int > mloc( Event::Mouse.Location.X() - X(), H() - ( Event::Mouse.Location.Y() - Y() ) + ( TopWindow::Display().H() - Bottom() ) );
	
	glReadPixels( mloc.X(), mloc.Y(), 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth );
	
	gluUnProject( mloc.X(), mloc.Y(), depth,
	              model_matrix, _projection_matrix, _viewport, &projected_x,
	              &projected_y, &projected_z );

	mesh._current_selection = Point3D( projected_x, projected_y, projected_z );

	// Find Mesh::_selection_radius
	gluUnProject( mloc.X() + MainBrush->GetRadius(), mloc.Y(), depth,
	              model_matrix, _projection_matrix, _viewport, &projected_x,
	              &projected_y, &projected_z );
	mesh._selection_radius = mesh._current_selection.Distance( Point3D( projected_x, projected_y, projected_z ) );
	//std::cout << mesh._selection_radius << std::endl;

	// Find symmetry points
	_symm_locs.clear();
	if( MainBrush->GetSymmetryX() )
	{
		gluProject( -mesh._current_selection.X(), mesh._current_selection.Y(), mesh._current_selection.Z(),
			    model_matrix, _projection_matrix, _viewport, &projected_x, &projected_y, &projected_z );
		_symm_locs.push_back( Math::Point2D< float >( projected_x, projected_y ) );
	}
	if( MainBrush->GetSymmetryY() )
	{
		gluProject( mesh._current_selection.X(), -mesh._current_selection.Y(), mesh._current_selection.Z(),
			    model_matrix, _projection_matrix, _viewport, &projected_x, &projected_y, &projected_z );
		_symm_locs.push_back( Math::Point2D< float >( projected_x, projected_y ) );
	}
	if( MainBrush->GetSymmetryZ() )
	{
		gluProject( mesh._current_selection.X(), mesh._current_selection.Y(), -mesh._current_selection.Z(),
			    model_matrix, _projection_matrix, _viewport, &projected_x, &projected_y, &projected_z );
		_symm_locs.push_back( Math::Point2D< float >( projected_x, projected_y ) );
	}
	if( MainBrush->GetSymmetryX() && MainBrush->GetSymmetryY() )
	{
		gluProject( -mesh._current_selection.X(), -mesh._current_selection.Y(), mesh._current_selection.Z(),
			    model_matrix, _projection_matrix, _viewport, &projected_x, &projected_y, &projected_z );
		_symm_locs.push_back( Math::Point2D< float >( projected_x, projected_y ) );
	}
	if( MainBrush->GetSymmetryX() && MainBrush->GetSymmetryZ() )
	{
		gluProject( -mesh._current_selection.X(), mesh._current_selection.Y(), -mesh._current_selection.Z(),
			    model_matrix, _projection_matrix, _viewport, &projected_x, &projected_y, &projected_z );
		_symm_locs.push_back( Math::Point2D< float >( projected_x, projected_y ) );
	}
	if( MainBrush->GetSymmetryY() && MainBrush->GetSymmetryZ() )
	{
		gluProject( mesh._current_selection.X(), -mesh._current_selection.Y(), -mesh._current_selection.Z(),
			    model_matrix, _projection_matrix, _viewport, &projected_x, &projected_y, &projected_z );
		_symm_locs.push_back( Math::Point2D< float >( projected_x, projected_y ) );
	}
	if( MainBrush->GetSymmetryX() && MainBrush->GetSymmetryY() && MainBrush->GetSymmetryZ() )
	{
		gluProject( -mesh._current_selection.X(), -mesh._current_selection.Y(), -mesh._current_selection.Z(),
			    model_matrix, _projection_matrix, _viewport, &projected_x, &projected_y, &projected_z );
		_symm_locs.push_back( Math::Point2D< float >( projected_x, projected_y ) );
	}

	// This code finds and stores the projected 2D locations
	// of all the mesh's vertices. Rather slow.
	//if( _recalc_projected ) // _Maybe_ someday...
	/*{
		std::cout << "recalcing proj..." << std::endl;
		if( _projected_points.size() != mesh._vertex_locations.size() || !_only_recalc_damaged )
		{
			std::cout << "recalcing all..." << std::endl;
			_projected_points.resize( mesh._vertex_locations.size() );
			for( int i = 0; i < mesh._vertex_locations.size(); i++ )
			{
				gluProject( mesh._vertex_locations[ i ].X(),
							mesh._vertex_locations[ i ].Y(),
							mesh._vertex_locations[ i ].Z(),
							model_matrix, _projection_matrix, _viewport,
							&projected_x, &projected_y, &projected_z );
				_projected_points[ i ] = Point3D( projected_x, projected_y, projected_z );
			}
		}
		if( _only_recalc_damaged )
		{
			std::cout << "recalcing damaged..." << std::endl;
			for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); i++ )
			{
				gluProject( mesh._vertex_locations[ mesh._modified_vertex_indices[ i ] ].X(),
							mesh._vertex_locations[ mesh._modified_vertex_indices[ i ] ].Y(),
							mesh._vertex_locations[ mesh._modified_vertex_indices[ i ] ].Z(),
							model_matrix, _projection_matrix, _viewport,
							&projected_x, &projected_y, &projected_z );
				_projected_points[ mesh._modified_vertex_indices[ i ] ] = Point3D( projected_x, projected_y, projected_z );
			}
			
			_only_recalc_damaged = false;
		}
		else
			std::cout << "NO RECALC!!!" << std::endl;
		_recalc_projected = false;
	}*/
}

void ViewGL::_mouse_down()
{
	if( Event::Mouse.LeftButtonDown /*&& !Fl::event_shift()*/ )
	{
		_history->AddMesh(); // Bottleneck!
		if( MainBrush->GetModelingMode() == Brush::GRAB )
		{
			_history->GetCurrentMesh().StartGrab();
		}
		ChangedMeshSignal();
		_update_vertex_buffer = false;
	}
	else if( Event::Mouse.RightButtonDown )
	{
		double xnorm = _last_mouse_location.X();
		double ynorm = _last_mouse_location.Y();
		ArcBall::NormalizeCoordinates( xnorm, ynorm, W(), H() );
		_arc_ball.BeginDrag( xnorm, ynorm );
	}
}

void ViewGL::_mouse_up()
{
	if( Event::Mouse.LeftButtonDown && MainBrush->GetModelingMode() == Brush::GRAB )
		_history->GetCurrentMesh().EndGrab();
	_store_view = true;
	_display_copied_view = true;
	_rendered = false;
	_recalc_projected = true;
	if( Event::Keyboard.ShiftDown && Event::Mouse.RightButtonDown )
		_arc_ball.BakeSnapRotation();
	Refresh();
}

void ViewGL::_mouse_move()
{
	//Mesh& mesh = _history->GetCurrentMesh();
	
	//_simple_initialize();
	//_initialize();
	//_find_view_normals();
	//_find_projected_points();
	if( Event::Mouse.MiddleButtonDown )
	{
		//_location.X() += ( Fl::event_x() - _x_last ) / 100.0;
		//_location.Y() += ( H() - Fl::event_y() - 1 - _y_last ) / 100.0;
		
		_location.X() += ( Event::Mouse.Location.X() - _last_mouse_location.X() ) / 100.0f;
		_location.Y() += ( Event::Mouse.Location.Y() - _last_mouse_location.Y() ) / -100.0f;
		_rendered = false;
	}
	else if( Event::Mouse.LeftButtonDown )
	{
		_update_vertex_buffer = true;

		bool changed = false;
		if( Event::Keyboard.ShiftDown )
		{
			MainBrush->SetStrength( -MainBrush->GetStrength() );
			changed = true;
		}
		//mesh._modified_vertex_indices.clear();
		if( MainBrush->GetModelingMode() == Brush::DISPLACE )
		{
			_history->GetCurrentMesh().DisplaceAreas();
			
			/*for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); i++ )
			{
				_damaged_polygon_indices.insert( _damaged_polygon_indices.begin(),
					mesh._vertex_polygon_indices[
					mesh._modified_vertex_indices[ i ] ].begin(),
					mesh._vertex_polygon_indices[
					mesh._modified_vertex_indices[ i ] ].end() );
			}
			_damaged_polygon_indices.unique();*/
			/*if( mesh._modified_vertex_indices.size() > 0 )
			{//x1,y1,x2,y2
				_damaged_area.resize( 4 );
				_damaged_area[ 0 ] = W();
				_damaged_area[ 1 ] = H();
				_damaged_area[ 2 ] = 0;
				_damaged_area[ 3 ] = 0;
				for( unsigned i = 0; i < mesh._modified_vertex_indices.size(); ++i )
				{
					Point3D& p = _projected_points[ mesh._modified_vertex_indices[ i ] ];
					if( p.X() < _damaged_area[ 0 ] )
						_damaged_area[ 0 ] = p.X();
					if( p.Y() < _damaged_area[ 1 ] )
						_damaged_area[ 1 ] = p.Y();
					if( p.X() > _damaged_area[ 2 ] )
						_damaged_area[ 2 ] = p.X();
					if( p.Y() > _damaged_area[ 3 ] )
						_damaged_area[ 3 ] = p.Y();
				}
				
			}*/
			//_rendered = false;
			
			//if( MainBrush->GetColorEdit() )
			/*if( _texture != NULL )
				UpdateTexture( _texture );*/
		}
		else if( MainBrush->GetModelingMode() == Brush::GRAB )
		{
			_history->GetCurrentMesh().ContinueGrab( ( Event::Mouse.Location.X() -
														_last_mouse_location.X() ) / 1000.0,
													 ( Event::Mouse.Location.Y() -
														_last_mouse_location.Y() ) / 1000.0 );
			_rendered = false;
		}
		else if( MainBrush->GetModelingMode() == Brush::SMOOTH )
			_history->GetCurrentMesh().SmoothAreas();
		else if( MainBrush->GetModelingMode() == Brush::PINCH )
			_history->GetCurrentMesh().PinchAreas();
		if( changed )
			MainBrush->SetStrength( -MainBrush->GetStrength() );
	}
	else if( Event::Mouse.RightButtonDown )
	{
		double xnorm = _last_mouse_location.X();
		double ynorm = _last_mouse_location.Y();
		ArcBall::NormalizeCoordinates( xnorm, ynorm, W(), H() );
		_arc_ball.Drag( xnorm, ynorm );
		_rendered = false;
		_store_view = false;
		_display_copied_view = false;
	}
	_last_mouse_location = Event::Mouse.Location;
}

void ViewGL::_scroll_down()
{
	_location.Z() -= _cam_distance / 10;
	_rendered = false;
}

void ViewGL::_scroll_up()
{
	_location.Z() += _cam_distance / 10;
	_rendered = false;
}
